import pandas as pd
import numpy as np
from functools import reduce
import argparse

from sklearn.pipeline import make_pipeline
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import StandardScaler

from emodpy_hiv.demographics.year_age_rate import YearAgeRate
import emodpy_hiv.demographics.year_age_rate as year_age_rate
import emodpy_hiv.demographics.un_world_pop as unwp


def infer_natural_mortality(year_age_rate_data: YearAgeRate,
                            interval_fit: tuple[float, float] = None,
                            predict_horizon: float = 2100.0) -> YearAgeRate:
    """
    Convert the raw mortality data to what would have been the expected natural mortality had HIV
    not caused a spike in mortality.  This algorithm assumes that during the 'interval_fit'
    mortality rates had a relatively steady decline (i.e. less people were dying each year),
    but shortly after the 'interval_fit' there was a spike in mortality.  The algorithm uses
    the data in the 'interval_fit' and extrapolates it out over the time period.

    It assumes that the input data is for only one node.

    Args:
        year_age_rate_data: This is a YearAgeRate data object containing the raw mortality data,
            probably output from the **extract_mortality()** function
        interval_fit: This tuple contains the range of years that we want to determine the
            mortality trend before the HIV epidemic.  These years will be extrapolated
            from the end of the fit forward.
        predict_horizon: This determines how far out the data will be extrapolated.

    Returns:
        A YearAgeRate object with the inferred mortality rates.
    """

    if len(year_age_rate_data.df[YearAgeRate.COL_NAME_NODE_ID].unique().tolist()) != 1:
        node_id_list = year_age_rate_data.df[YearAgeRate.COL_NAME_NODE_ID].unique().tolist()
        msg  =  "Invalid number of nodes in the YearAgeRate object.\n"         # noqa: E221, E222
        msg +=  "The algorithm currently only supports one node at a time.\n"  # noqa: E222
        msg += f"The input has the following nodes: {node_id_list}"
        raise ValueError(msg)

    if interval_fit is None:
        interval_fit = (1970, 1980)
    elif interval_fit[1] <= interval_fit[0]:
        msg  =  "Invalid 'interval_fit' values.\n" # noqa: E221, E222
        msg += f"The first value in the tuple, {interval_fit[0]} must be strictly less\n"
        msg += f"than the second value, {interval_fit[1]}."
        raise ValueError(msg)

    if predict_horizon <= interval_fit[1]:
        msg  = "Invalid value of 'predict_horizon' with respect to 'interval_fit'.\n" # noqa: E221
        msg += f"The 'predicted_horizon' (={predict_horizon}) must be greater than\n"
        msg += f"the second value of the 'interval_fit[1]' (={interval_fit[1]})."

    df_mort = year_age_rate_data.df.copy()

    # ---------------------------------------------------
    # --- Log transform the data and sort by year and age
    # ---------------------------------------------------
    df_mort[YearAgeRate.COL_NAME_RATE] = df_mort[YearAgeRate.COL_NAME_RATE].apply(lambda x: np.log(x))
    df_mort.sort_values([YearAgeRate.COL_NAME_MIN_YEAR, YearAgeRate.COL_NAME_MIN_AGE], inplace=True)

    # -------------------------------------------------------------------
    # --- Extract the data for that exists BEFORE the reference interval.
    # --- Do not include any data from the interval
    # --- inclusive=left means the right is not inclusive.
    # -------------------------------------------------------------------
    min_year_col = df_mort[YearAgeRate.COL_NAME_MIN_YEAR]
    df_before_time = df_mort[min_year_col.between(0, interval_fit[0], inclusive='left')].copy()

    # -------------------------------------------------------------------------
    # --- Get list of ages to extrapolate the data for and setup the dataframe
    # -------------------------------------------------------------------------
    age_list = df_mort[YearAgeRate.COL_NAME_MIN_AGE].unique().tolist()
    df_mort.set_index([YearAgeRate.COL_NAME_MIN_AGE], inplace=True)

    df_list = []
    for age in age_list:
        tmp_data = df_mort.loc[age, :]
        extrap_model = make_pipeline(StandardScaler(with_mean=False), LinearRegression())

        # -------------------------------------------------------------------------------------------------
        # --- Get the data (log(rate)) for the reference interval and from the beginning of the reference
        # --- interval all the way to the end of the prediction.
        # -------------------------------------------------------------------------------------------------
        min_year_col = tmp_data[YearAgeRate.COL_NAME_MIN_YEAR]
        first_extrap_df = tmp_data[min_year_col.between(interval_fit[0], interval_fit[1])]
        xx              = tmp_data[min_year_col.between(interval_fit[0], predict_horizon)].values[:, 1] # noqa: E221

        # ----------------------------------------------------------------------------------------
        # --- Have the model fit the reference interval and then use that to predict the interval
        # --- from the beginning of the reference interval to the end.
        # ----------------------------------------------------------------------------------------
        values = first_extrap_df.values
        extrap_model.fit(values[:, 1].reshape(-1, 1), values[:, 2])
        extrap_predictions = extrap_model.predict(xx.reshape(-1, 1))

        # ----------------------------------------------------
        # --- Create a new dataframe with the new predictions
        # ----------------------------------------------------
        extrap_dict = {
            YearAgeRate.COL_NAME_MIN_AGE: age,
            YearAgeRate.COL_NAME_MIN_YEAR: xx,
            'Extrap': extrap_predictions
        }
        loc_df = pd.DataFrame.from_dict(extrap_dict)
        loc_df.set_index([YearAgeRate.COL_NAME_MIN_AGE, YearAgeRate.COL_NAME_MIN_YEAR], inplace=True)

        df_list.append(loc_df.copy())

    # ---------------------------------------------------------------
    # --- Concatenate the dataframes from the different ages together
    # ---------------------------------------------------------------
    df_e1 = pd.concat(df_list, axis=0)

    # -----------------------------------------------------------
    # --- Merge this datafame with the predictions with original
    # -----------------------------------------------------------
    df_list_final = [df_mort, df_e1]
    df_total = reduce(lambda left, right: pd.merge(left, right, on=[YearAgeRate.COL_NAME_MIN_AGE, YearAgeRate.COL_NAME_MIN_YEAR]), df_list_final)
    df_total = df_total.reset_index(inplace=False).set_index([YearAgeRate.COL_NAME_MIN_AGE], inplace=False)

    # ----------------------------------------------
    # --- Convert the data to rates from log(rates)
    # ----------------------------------------------
    df_total['Extrap'] = df_total['Extrap'].apply(np.exp)
    df_total['Data'] = df_total[YearAgeRate.COL_NAME_RATE].apply(np.exp)

    # ------------------------------------------------------------------
    # --- Convert the before time data to rates from log(rates) and
    # --- combine the data into one dataframe for the entire time period
    # ------------------------------------------------------------------
    df_before_time['Data'] = df_before_time[YearAgeRate.COL_NAME_RATE].apply(np.exp)
    df_before_time.set_index([YearAgeRate.COL_NAME_MIN_AGE], inplace=True)
    df_total = pd.concat([df_total, df_before_time], axis=0, join='outer', sort=True)
    df_total.reset_index(inplace=True)
    df_total.sort_values(by=[YearAgeRate.COL_NAME_MIN_YEAR, YearAgeRate.COL_NAME_MIN_AGE], inplace=True)

    # ----------------
    # --- Replace NaNs
    # ----------------
    def min_not_nan(x_list):
        loc_in = list(filter(lambda x: not np.isnan(x), x_list))
        return np.min(loc_in)
    df_total[YearAgeRate.COL_NAME_RATE] = df_total[['Data', 'Extrap']].apply(min_not_nan, axis=1)

    # ------------------------------------------------------
    # --- Convert dataframe to one suitable for YearAgeRate
    # ------------------------------------------------------
    df_total.drop(columns=['Data', 'Extrap'], inplace=True)
    df_total = df_total[YearAgeRate.COL_NAMES]
    df_total.sort_values(by=YearAgeRate.SORT_BY_COLUMNS, inplace=True)

    return YearAgeRate(df=df_total)


def mortality_read_infer_plot(country: str,
                              version: str,
                              gender: str,
                              interval_fit: tuple[float, float] = None,
                              save_data: bool = False,
                              other_csv_filename: str = None,
                              img_dir: str = None,
                              filename_to_save_to: str = None) -> None:
    """
    Extract the mortality data from the given file for the given country and plot
    both the raw data and the inferred data (without HIV deaths).  The plot window will
    have a plot for each age where the data in the plot shows the rate per year for
    that age group.  The label shows the minimum age of the band.  This means that the
    age range being represented is from this value to the next largest age that is plotted.

    Args:
        un_world_pop_filename: The path to a UN World Population spreadsheet containing
            mortality data.  It can be for either female or male.

        country: The name of the country to be extracted.  It must match exactly to the
            country name used in the referenced spreadsheet.

        version: The year/version of the indicated spreadsheet.  Each year has a slightly
            different format.  Supported versions are 2012, 2015, 2019, 2024

        gender:
            The gender of the data to be extracted.  Possible values are 'male' and 'female'.

        save_data: If true (default is False), two YearAgeRate CSV formatted files are
            output into the current directory.  The files are: 'raw_mortality_year_age_rate.csv'
            and 'mortality_minus_hiv_year_age_rate.csv'.

        other_csv_filename: If the filename is defined, it reads an expected YearAgeRate CSV
            file and plots it with the raw and inferred data from the spreadsheet.
            This allows the user to plot other mortality data (i.e. older versions)
            with the new data.

        img_dir: If this is defined, the images are saved to this directory.  If not defined,
            the images are displayed in a window.

        filename_to_save_to: The name of the file to save the image to.  This is only used if
            img_dir is defined.

    Return:
        No return
    """
    raw_mortality_yar = unwp.extract_mortality(country=country,
                                               version=version,
                                               gender=gender)
    mortality_minus_hiv_yar = infer_natural_mortality(year_age_rate_data=raw_mortality_yar,
                                                      interval_fit=interval_fit,
                                                      predict_horizon=2100)

    if save_data:
        raw_mortality_yar.to_csv("raw_mortality_year_age_rate.csv")
        mortality_minus_hiv_yar.to_csv("mortality_minus_hiv_year_age_rate.csv")

    plot_list = []
    plot_list.append(raw_mortality_yar)
    plot_list.append(mortality_minus_hiv_yar)

    if other_csv_filename:
        other_yar = YearAgeRate(csv_filename=other_csv_filename)
        plot_list.append(other_yar)

    title = f"Red: UN World Population, country={country}, version={version}, gender={gender}\n"
    title += f"Blue: Natural mortality inferred from {interval_fit[0]} to {interval_fit[1]}"
    year_age_rate.plot(year_age_rate_list=plot_list,
                       title=title,
                       node_id=0,
                       img_dir=img_dir,
                       filename_to_save_to=filename_to_save_to)


def mortality_read_infer_plot_app(country: str,
                                  version: str,
                                  gender: str,
                                  min_year: int,
                                  max_year: int,
                                  save_data: bool = False,
                                  other_csv_filename: str = None,
                                  img_dir: str = None) -> None:
    """
    Using the 'version' and 'gender', select the UN World Population mortality data file.
    Extract the mortality data from the selected file for the given country and plot
    both the raw data and the inferred data (without HIV deaths).  The plot will
    have a plot for each age where the data in the plot shows the rate per year for
    that age group.  The label shows the minimum age of the band.  This means that the
    age range being represented is from this value to the next largest age that is plotted.

    Args:
        country: The name of the country to be extracted.  It must match exactly to the
            country name used in the referenced spreadsheet.

        version: The year/version of the indicated spreadsheet.  Each year has a slightly
            different format.  Supported versions are 2012, 2015, 2019, 2024

        gender: The gender of the data to be extracted.  Possible values are 'male' and 'female'.

        min_year: The start year of the interval to fit the data to.

        max_year: The end year of the interval to fit the data to.

        save_data: If true (default is False), two YearAgeRate CSV formatted files are
            output into the current directory.  The files are: 'raw_mortality_year_age_rate.csv'
            and 'mortality_minus_hiv_year_age_rate.csv'.

        other_csv_filename: If the filename is defined, it reads an expected YearAgeRate CSV
            file and plots it with the raw and inferred data from the spreadsheet.
            This allows the user to plot other mortality data (i.e. older versions)
            with the new data.

        img_dir: If this is defined, the images are saved to this directory.  If not defined,
            the images are displayed in a window.

    Return:
        No return
    """
    if min_year >= max_year:
        raise ValueError(f"Invalid interval: {min_year} >= {max_year}.\n"
                         + "The min year must be less than the max year.")
    if min_year < 1950:
        raise ValueError(f"Invalid interval: {min_year} < 1950.\n"
                         + "The min year must be greater than or equal to 1950,\n"
                         + "because the data is not present in the files.")
    if max_year > 2100:
        raise ValueError(f"Invalid interval: {max_year} > 2100.\n"
                         + "The max year must be less than or equal to 2100,\n"
                         + "because the data is not present in the files.")

    filename_to_save_to = f"Inferred_Mortality_{gender}_{country}_{version}_from_{min_year}_to_{max_year}.png"
    filename_to_save_to = filename_to_save_to.replace(" ", "_")

    mortality_read_infer_plot(country=country,
                              version=version,
                              gender=gender,
                              interval_fit=(min_year, max_year),
                              save_data=save_data,
                              other_csv_filename=other_csv_filename,
                              img_dir=img_dir,
                              filename_to_save_to=filename_to_save_to)


if __name__ == '__main__':

    parser = argparse.ArgumentParser()

    parser.add_argument('-g', '--gender', type=str, default='male', help="Gender of mortality data. Options: 'male' or 'female'")
    parser.add_argument('-y', '--year', type=str, default="2015", help="Year of UN World Population data.  Options: 2012, 2015, 2019, 2024")
    parser.add_argument('-c', '--country', type=str, default="Zambia", help="Country name in UN World Population data.")
    parser.add_argument('-m', '--min_year', type=int, default=1950, help="The start year of the interval to fit the data to.")
    parser.add_argument('-x', '--max_year', type=int, default=1975, help="The end year of the interval to fit the data to.")
    parser.add_argument('-o', '--output', type=str, default=None, help='If provided, a directory will be created and images saved to the folder.  If not provided, it opens windows.')

    args = parser.parse_args()

    mortality_read_infer_plot_app(country=args.country,
                                  version=args.year,
                                  gender=args.gender,
                                  min_year=args.min_year,
                                  max_year=args.max_year,
                                  save_data=False,
                                  other_csv_filename=None,
                                  img_dir=args.output)



# ported from: emodpy/campaign/node_intervention.py
# These are the node-level intervention classes that are available in this package, please import them from this module.
from emodpy.campaign.node_intervention import MultiNodeInterventionDistributor
from emodpy.campaign.node_intervention import BroadcastNodeEvent
from emodpy.campaign.node_intervention import ImportPressure
from emodpy.campaign.node_intervention import MigrateFamily
from emodpy.campaign.node_intervention import NodePropertyValueChanger
from emodpy.campaign.node_intervention import Outbreak


# __all_exports: A list of classes that are intended to be exported from this module.
# the private classes are commented out until we have time to review and test them.
__all_exports = [
    MultiNodeInterventionDistributor,
    BroadcastNodeEvent,
    ImportPressure,
    MigrateFamily,
    NodePropertyValueChanger,
    Outbreak
]

# The following loop sets the __module__ attribute of each class in __all_exports to the name of the current module.
# This is done to ensure that when these classes are imported from this module, their __module__ attribute correctly
# reflects their source module.

for _ in __all_exports:
    _.__module__ = __name__

# __all__: A list that defines the public interface of this module.
# This is essential to ensure that Sphinx builds documentation for these classes, including those that are imported
# from emodpy.
# It contains the names of all the classes that should be accessible when this module is imported using the syntax
# 'from module import *'.
# Here, it is set to the names of all classes in __all_exports.

__all__ = [_.__name__ for _ in __all_exports]



from enum import Enum

from emodpy.utils.targeting_config import AbstractTargetingConfig, BaseTargetingConfig
from emodpy.utils.targeting_config import HasIP, HasIntervention, IsPregnant


class YesNoNa(Enum):
    """
    This enum is used in the Targeting_Config logic to indicate if a particular attribute
    should be yes, no, or na (not applicable/don't consider)
    """
    YES = "YES"
    NO  = "NO"    # noqa: E221
    NA  = "NA"    # noqa: E221


class MoreOrLess(Enum):
    """
    This enum is used in Targeting_Config logic to indicate if the check should be
    strictly less than, '<', or strictly greater than, '>'.  These are used when
    checking for the number of partners, months, etc.
    """
    LESS = "LESS" # less than '<'
    MORE = "MORE" # greater than '>'


class OfRelationshipType(Enum):
    """
    This enum is used in the Targeting_Config logic to indicate if the check should
    consider specific relationship types or not.  If the user selects a specific
    type, then the rest of the checks will only be for those types of relationships.
    If the user selects 'NA', then relationships of all types will be considered.
    """
    NA         = "NA"           # noqa: E221
    TRANSITORY = "TRANSITORY"
    INFORMAL   = "INFORMAL"     # noqa: E221
    MARITAL    = "MARITAL"      # noqa: E221
    COMMERCIAL = "COMMERCIAL"


class NumMonthsType(Enum):
    """
    This enum is used in the Targeting_Config logic to indicate the duration range
    to check how many partners a person has.  These ranges are fixed in EMOD and
    the user cannot indicate a random range.
    """
    THREE_MONTHS  = "THREE_MONTHS"   # noqa: E221
    SIX_MONTHS    = "SIX_MONTHS"     # noqa: E221
    NINE_MONTHS   = "NINE_MONTHS"    # noqa: E221
    TWELVE_MONTHS = "TWELVE_MONTHS"


class RecentlyType(Enum):
    """
    This enum is used in the Targeting_Config logic to indicate if we only want
    to consider people who have a relationship that just started or just ended.
    This means we are interested in relationships that started or stopped during
    the current time step.
    """
    NA      = "NA"       # noqa: E221
    STARTED = "STARTED"
    ENDED   = "ENDED"    # noqa: E221


class RelationshipTerminationReasonType(Enum):
    """
    This enum is used in the Targeting_Config logic to indicate if we are
    interested in a person because of WHY a relationship recently ended.
    For example, if a person BROKE_UP and has not other partners, we might
    want them to stop using PrEP.
    """

    NA = "NA"
    """
    The relationship has not been terminated.
    """

    BROKE_UP = "BROKEUP" # EMOD does not have the underscore
    """
    The relationship ended due to the duration settings.
    """

    SELF_MIGRATING = "SELF_MIGRATING"
    """
    One of the partners in the relationship has decided to migrate and
    so the relationship is terminated. Note: The user can control what
    happens to a relationship when there is migration; it does not have
    to terminate.
    """

    PARTNER_DIED = "PARTNER_DIED"
    """
    One of the partners died so the relationship was terminated.
    """

    PARTNER_TERMINATED = "PARTNER_TERMINATED"
    """
    This happens when the couple is separated due to migration and one
    of the partners decides to terminate the relationship.
    """

    PARTNER_MIGRATING = "PARTNER_MIGRATING"
    """
    The relationship is being terminated because one of the partners has
    another partner that is migrating. For example, a married couple is
    moving because the wife got a new job. The husband must terminate his
    other relationships with "PARTNER_MIGRATING".
    """


class IsCircumcised(BaseTargetingConfig):
    """
    Select the individual based on whether or not they are circumcised.
    """
    def __init__(self):
        super().__init__()
        self.class_name = "IsCircumcised"


class IsHivPositive(BaseTargetingConfig):
    """
    Select the individual based on whether or not they have HIV.  The "and_has_XXX"
    parameters extend this by being and'd with the check on whether or not the person
    has HIV.  For example, if you want to select people that are:

        * HIV negative
        * AND have been tested
        * AND have never tested positive

    you will create the following:

        >>> targeting_config = ~IsHivPositive( and_has_ever_been_tested=YesNoNa.YES,
        >>>                                    and_has_ever_tested_positive=YesNoNa.NO )

    Notice that if the person being considered was HIV positive, the rest of the checks
    would not matter because the first check was false.

    Also notice that when you invert IsHivPositive, you are only changing the whether
    or not you are looking for infected people.  It does not impact the "and_has_XXX" checks.

    Args:
        and_has_ever_been_tested: If the user sets this Enum to 'YES', then the
            individual's true infection status must equal to the inversion status and the
            person must have been tested at least once. Notice that this only tells if the
            has been tested, NOT that they tested positive. If set to 'NA' (default), then
            do not include this as part of the check.

        and_has_ever_tested_positive: If the user sets this Enum to 'YES', then the
            individual's true infection status must equal to the inversion status and the
            person must have tested POSITIVE at least once. Notice that this is different
            than just having been tested. However, it does not say the person received the
            results. If set to 'NA' (default), then do not include this as part of the check.

        and_has_received_positive_results: If the user sets this Enum to 'YES', then the
            individual's true infection status must equal to the inversion status and the
            last test result received was positive. If set to 'NA' (default), then do not
            include this as part of the check.
    """
    def __init__(self,
                 and_has_ever_been_tested:          YesNoNa = YesNoNa.NA,   # noqa: E241
                 and_has_ever_tested_positive:      YesNoNa = YesNoNa.NA,   # noqa: E241
                 and_has_received_positive_results: YesNoNa = YesNoNa.NA):
        super().__init__()
        self.class_name = "IsHivPositive"
        self.been_tested = and_has_ever_been_tested
        self.tested_positive = and_has_ever_tested_positive
        self.received_positive_results = and_has_received_positive_results

    def to_schema_dict(self, campaign):
        """
        Create the ReadOnlyDict object representation of this Targeting_Config logic.
        This is the dictionary used to generate the JSON for EMOD.

        Args:
            campaign: The campaign module that has the path to the schema

        Returns:
            A ReadOnlyDict object created by schema_to_class
        """
        tc_obj = super().to_schema_dict(campaign)
        tc_obj.And_Has_Ever_Been_Tested          = self.been_tested.value                # noqa: E221
        tc_obj.And_Has_Ever_Tested_Positive      = self.tested_positive.value            # noqa: E221
        tc_obj.And_Has_Received_Positive_Results = self.received_positive_results.value
        return tc_obj


class IsOnART(BaseTargetingConfig):
    """
    Select the individual based on whether or not they are actively on ART.
    """
    def __init__(self):
        super().__init__()
        self.class_name = "IsOnART"


class IsPostDebut(BaseTargetingConfig):
    """
    Select the individual based on whether or not they have reached sexual debut.
    """
    def __init__(self):
        super().__init__()
        self.class_name = "IsPostDebut"


class HasBeenOnArtMoreOrLessThanNumMonths(BaseTargetingConfig):
    """
    Determine if the individual has been on ART for less than "num" months.
    It will be false if the individual is not on ART. The test will be
    strictly less than.

    Args:
        num_months: This is the number of months that will be used in the test.
            The individual's duration on ART must be more or less than this value.
            NOTE: 1500 months is about 12 months / year * 125 years of max age

        more_or_less: This is used to determine if the check should be less than
            or greater than.
    """
    def __init__(self,
                 num_months: float = 1500.0,
                 more_or_less: MoreOrLess = MoreOrLess.LESS):
        super().__init__()
        self.class_name = "HasBeenOnArtMoreOrLessThanNumMonths"
        # TODO: use Ye's validate range check on num_months
        self.num_months = num_months
        self.more_or_less = more_or_less

    def to_schema_dict(self, campaign):
        """
        Create the ReadOnlyDict object representation of this Targeting_Config logic.
        This is the dictionary used to generate the JSON for EMOD.

        Args:
            campaign: The campaign module that has the path to the schema

        Returns:
            A ReadOnlyDict object created by schema_to_class
        """
        tc_obj = super().to_schema_dict(campaign)
        tc_obj.Num_Months   = self.num_months          # noqa: E221
        tc_obj.More_Or_Less = self.more_or_less.value
        return tc_obj


class HasMoreOrLessThanNumPartners(BaseTargetingConfig):
    """
    Determine if the individual has more or less than a specified number of active partners.
    This includes relationships that are paused due to migration. This test is strictly more
    or less than.

    Args:
        num_partners: This parameter allows the user to set the number of active
            partners/relationships that the individual must have more or less of.
            The value can range between 0 and 62.

        more_or_less: This is used to determine if the check should be less than
            or greater than.  The default value is 'LESS'.

        of_relationship_type: If the user sets this value to one of the four specific types
            (TRANSITORY, INFORMAL, MARITAL, COMMERCIAL), then the individual must have more or
            less relationships of this type. When the value is set to 'NA' (default), then it
            will count the relationships regardless of type.
    """
    def __init__(self,
                 num_partners: int = 0,
                 more_or_less: MoreOrLess = MoreOrLess.LESS,
                 of_relationship_type: OfRelationshipType = OfRelationshipType.NA):
        super().__init__()
        self.class_name = "HasMoreOrLessThanNumPartners"

        # TODO: use Ye's validate range method on num_partners
        self.num_partners = num_partners
        self.more_or_less = more_or_less
        self.of_relationship_type = of_relationship_type

    def to_schema_dict(self, campaign):
        """
        Create the ReadOnlyDict object representation of this Targeting_Config logic.
        This is the dictionary used to generate the JSON for EMOD.

        Args:
            campaign: The campaign module that has the path to the schema

        Returns:
            A ReadOnlyDict object created by schema_to_class
        """
        tc_obj = super().to_schema_dict(campaign)
        tc_obj.Num_Partners         = self.num_partners               # noqa: E221
        tc_obj.More_Or_Less         = self.more_or_less.value         # noqa: E221
        tc_obj.Of_Relationship_Type = self.of_relationship_type.value
        return tc_obj


class HasHadMultiplePartnersInLastNumMonths(BaseTargetingConfig):
    """
    Determine if the individual has had more than one relationship in the last "Num" months.
    This could constitute as "high-risk" behavior. The goal is to target people that have had
    coital acts with more than one person during the last X months. This would count current
    relationships, relationships that started in the last X months, and relationships that
    have ended in the last X months. Basically, all the relationships that have been active
    at some point during the last X months.

    NOTE: This only counts unique partners. Two relationships with the same person during
    the time period will count as one. For example, if you dated a person for three months,
    broke up, got back together after six months, and the time period was twelve months, then
    it would be considered one partner.

    NOTE: Also note that one partner can count across multiple periods.  For example, if the
    person has been in a MARITAL relationship for years, that partner will be counted in the
    last three months, last six months, last nine months, and last twelve months.  If the
    person started a relationship last month, it will only be counted within the last three
    months.

    Args:
        num_month_type: This parameter allows the user to set the maximum number of months
            from the current day to consider if the individual had multiple relationships.

        of_relationship_type: If the user sets this value to one of the four specific types
            (TRANSITORY, INFORMAL, MARITAL, COMMERCIAL), then the individual must have had
            more than one relationship of this type. When the value is set to 'NA' (default),
            then it just matters if the person had more than one relationship, regardless
            of type.
    """
    def __init__(self,
                 num_month_type: NumMonthsType = NumMonthsType.THREE_MONTHS,
                 of_relationship_type: OfRelationshipType = OfRelationshipType.NA):
        super().__init__()
        self.class_name = "HasHadMultiplePartnersInLastNumMonths"
        self.num_month_type = num_month_type
        self.of_relationship_type = of_relationship_type

    def to_schema_dict(self, campaign):
        """
        Create the ReadOnlyDict object representation of this Targeting_Config logic.
        This is the dictionary used to generate the JSON for EMOD.

        Args:
            campaign: The campaign module that has the path to the schema

        Returns:
            A ReadOnlyDict object created by schema_to_class
        """
        tc_obj = super().to_schema_dict(campaign)
        tc_obj.Num_Months_Type      = self.num_month_type.value       # noqa: E221
        tc_obj.Of_Relationship_Type = self.of_relationship_type.value
        return tc_obj


class HasCd4BetweenMinAndMax(BaseTargetingConfig):
    """
    Determine if the individual has a CD4 count that is between "min" and "max".
    The test is inclusive for "min" and exclusive for "max".

    Args:
        min_cd4: The minimum value of the test range. The individuals CD4 can be
            equal to this value to be considered "between".  The value can range
            from 0 to 1999 and must be less than 'max_cd4'.

        max_cd4: The maximum value of the test range. The individual's CD4 must be
            strictly less than this value to be considered "between".  The value
            can range from 0 to 2000 and must be greater than 'min_cd4'.
    """
    def __init__(self, min_cd4: float = 0, max_cd4: float = 2000):
        super().__init__()
        self.class_name = "HasCd4BetweenMinAndMax"

        # TODO: Use Ye's validate value range method
        if min_cd4 >= max_cd4:
            msg = f"Invalid 'min_cd4'={min_cd4} and 'max_cd4'={max_cd4}.\n"
            msg += "min_cd4' must be strictly less than 'max_cd4'."
            raise ValueError(msg)

        self.min_cd4 = min_cd4
        self.max_cd4 = max_cd4

    def to_schema_dict(self, campaign):
        """
        Create the ReadOnlyDict object representation of this Targeting_Config logic.
        This is the dictionary used to generate the JSON for EMOD.

        Args:
            campaign: The campaign module that has the path to the schema

        Returns:
            A ReadOnlyDict object created by schema_to_class
        """
        tc_obj = super().to_schema_dict(campaign)
        tc_obj.Min_CD4 = self.min_cd4
        tc_obj.Max_CD4 = self.max_cd4
        return tc_obj


class HasRelationship(BaseTargetingConfig):
    """
    This is used to select people who have a partner/relationship that meets certain
    qualifications.

    Args:
        of_relationship_type: If the user sets this value to one of the four specific
            types (TRANSITORY, INFORMAL, MARITAL, COMMERCIAL), then the individual must
            have at least one relationships of this type.(default).  The other constraints
            will be on these relationships.  If NA is selected, then all of the individual's
            relationships will be considered.

        that_recently: This parameter is used if the relationship being considered must have
            recently been started or ended. Possible values are:

            - NA (Default) - Do no consider That_Recently in the selection process.
            - STARTED - Only consider relationships that have started within one time-step.
                One should note that the relationships you see will depend on whether you are
                using NodeLevelHealthTriggeredIV (NLHTIV) [add_intervention_triggered()] or an
                event coordinator [add_intervention_scheduled()]. NLHTIV and the individuals
                will be updated after new relationships have been created so with this feature,
                NLHTIV will see relationships in the previous time step as well as the current
                time step.  An event coordinator executes BEFORE new relationships are formed
                so it will only see the relationships created in the previous time step.
            - ENDED - Check the status of the relationship that just ended. Note: This can only
                be used with NodeLevelHealthTriggeredIV listening for the 'ExitedRelationship'
                event.

        that_recently_ended_due_to: If That_Recently is set to 'ENDED', this is used to look at the
            reason the relationship ended. Possible values are:

            - NA (Default) - The relationship has not been terminated.
            - BROKE_UP - The relationship ended due to the duration settings.
            - SELF_MIGRATING - One of the partners in the relationship has decided to migrate and so
                the relationship is terminated. Note: the user can control what happens to a
                relationship when there is migration; it does not have to terminate.
            - PARTNER_MIGRATING - The relationship is being terminated because one of the partners
                has another partner that is migrating. For example, a married couple is moving because
                the wife got a new job. The husband must terminate his other relationships with 'PARTNER_MIGRATING'.
            - PARTNER_DIED - One of the partners died so the relationship was terminated.
            - PARTNER_TERMINATED - This happens when the couple is separated due to migration and one
                of the partners decides to terminate the relationship.

        with_partner_who: Given that the parameters about the relationship are true, this parameter
            is used to look at the partner of the relationship. It is a Targeting_Config so one uses
            the same classes to query the partner. For example, to find out if person has a partner
            with HIV, you could use IsHivPositive.
    """
    def __init__(self,
                 of_relationship_type:       OfRelationshipType                = OfRelationshipType.NA,                 # noqa: E221, E241
                 that_recently:              RecentlyType                      = RecentlyType.NA,                       # noqa: E221, E241
                 that_recently_ended_due_to: RelationshipTerminationReasonType = RelationshipTerminationReasonType.NA,
                 with_partner_who:           AbstractTargetingConfig           = None):                                 # noqa: E221, E241
        super().__init__()
        self.class_name = "HasRelationship"

        if (of_relationship_type == OfRelationshipType.NA) and\
           (that_recently        == RecentlyType.NA      ) and\
           (with_partner_who     is None                 ):      # noqa: E202, E272, E221
            msg = "No parameters were configured so no relationships would be targeted.\n"
            msg += "'HasRelationship' requires that you define at least one parameter."
            raise ValueError(msg)

        self.of_relationship_type       = of_relationship_type       # noqa: E221
        self.that_recently              = that_recently              # noqa: E221
        self.that_recently_ended_due_to = that_recently_ended_due_to
        self.with_partner_who           = with_partner_who           # noqa: E221

    def to_schema_dict(self, campaign):
        """
        Create the ReadOnlyDict object representation of this Targeting_Config logic.
        This is the dictionary used to generate the JSON for EMOD.

        Args:
            campaign: The campaign module that has the path to the schema

        Returns:
            A ReadOnlyDict object created by schema_to_class
        """
        tc_obj = super().to_schema_dict(campaign)
        tc_obj.Of_Relationship_Type       = self.of_relationship_type.value        # noqa: E221
        tc_obj.That_Recently              = self.that_recently.value               # noqa: E221
        tc_obj.That_Recently_Ended_Due_To = self.that_recently_ended_due_to.value

        if self.with_partner_who:
            tc_obj.With_Partner_Who = self.with_partner_who.to_schema_dict(campaign)

        return tc_obj

    def to_simple_dict(self, campaign):
        """
        Return a plain/simple dictionary of the expected JSON for EMOD.  The main
        purpose of this is for validation in testing.  We need the ability to see
        that the logic written in python is translated to the JSON correctly.

        Args:
            campaign: The campaign module that has the path to the schema

        Returns:
            A simple dictionary containing the data for EMOD.
        """
        tc_dict = super().to_simple_dict(campaign)
        if self.with_partner_who:
            # we need to convert this object here because it is easier to ensure the object
            # is converted to the simple dictionary.
            tc_dict["With_Partner_Who"] = self.with_partner_who.to_simple_dict(campaign)
        else:
            # This gets rid of the default stuff put into the dictionary
            del tc_dict["With_Partner_Who"]
        return tc_dict


# __all_exports: A list of classes that are intended to be exported from this module.

__all_exports = [
    AbstractTargetingConfig,
    BaseTargetingConfig,
    HasIP,
    HasIntervention,
    IsPregnant,
    YesNoNa,
    MoreOrLess,
    OfRelationshipType,
    NumMonthsType,
    RecentlyType,
    RelationshipTerminationReasonType,
    IsCircumcised,
    IsHivPositive,
    IsOnART,
    IsPostDebut,
    HasBeenOnArtMoreOrLessThanNumMonths,
    HasMoreOrLessThanNumPartners,
    HasHadMultiplePartnersInLastNumMonths,
    HasCd4BetweenMinAndMax,
    HasRelationship
]

# The following loop sets the __module__ attribute of each class in __all_exports to the name of the current module.
# This is done to ensure that when these classes are imported from this module, their __module__ attribute correctly
# reflects their source module.
# During the documentation build with Sphinx, these classes will be displayed as belonging to the 'emodpy_hiv' package,
# not the 'emodpy' package.
# For example, the 'PropertyRestrictions' class will be documented as 'emodpy_hiv.campaign.common.PropertyRestrictions(...)'.
# This is crucial for accurately representing the source of these classes in the documentation.

for _ in __all_exports:
    _.__module__ = __name__

# __all__: A list that defines the public interface of this module.
# This is essential to ensure that Sphinx builds documentation for these classes, including those that are imported
# from emodpy.
# It contains the names of all the classes that should be accessible when this module is imported using the syntax
# 'from module import *'.
# Here, it is set to the names of all classes in __all_exports.

__all__ = [_.__name__ for _ in __all_exports]



import pandas as pd
import os
import matplotlib.pyplot as plt


def create_common_x_values(a_old, b_old):
    """
    Matplotlib needs the x-values to be the same for different (x,y) lines being plotted.
    This method creates a new version of each set of values.
    """
    a_new = []
    b_new = []

    a_index = 0
    b_index = 0
    while (a_index < len(a_old)) and (b_index < len(b_old)):
        a = a_old[a_index]
        b = b_old[b_index]
        if a < b:
            if b_index == 0:
                a_new.append(a_old[a_index])
                a_index = a_index + 1
            else:
                a_new.append(a_old[a_index])
                b_new.append(a_old[a_index])
                a_index = a_index + 1
        elif a == b:
            a_new.append(a_old[a_index])
            b_new.append(b_old[b_index])
            a_index = a_index + 1
            b_index = b_index + 1
        else:
            if a_index == 0:
                b_new.append(b_old[b_index])
                b_index = b_index + 1
            else:
                a_new.append(b_old[b_index])
                b_new.append(b_old[b_index])
                b_index = b_index + 1

    while a_index < len(a_old):
        a_new.append(a_old[a_index])
        a_index = a_index + 1

    while b_index < len(b_old):
        b_new.append(b_old[b_index])
        b_index = b_index + 1

    return a_new, b_new


def fill_in_y_values(x_old, y_old, x_new):
    """
    Since matplotlib needs the lines/curves to have the same X-values,
    we need to find the associated Y-values when we add the new X-values.
    This method uses linear interpolation to find the values between points.
    """
    y_new = []

    x_old_index = 0
    x_new_index = 0
    while (x_old_index < len(x_old)) and (x_new_index < len(x_new)):
        if x_old[x_old_index] == x_new[x_new_index]:
            y_new.append(y_old[x_old_index])
            x_old_index = x_old_index + 1
            x_new_index = x_new_index + 1
        elif x_old[x_old_index] > x_new[x_new_index]:
            x1 = x_old[x_old_index - 1]  # noqa: E201, E202
            x2 = x_old[x_old_index    ]  # noqa: E201, E202
            y1 = y_old[x_old_index - 1]  # noqa: E201, E202
            y2 = y_old[x_old_index    ]  # noqa: E201, E202
            x  = x_new[x_new_index    ]  # noqa: E201, E202, E221
            y  = y1 + (x - x1) * (y2 - y1) / (x2 - x1)   # noqa: E221
            y_new.append(y)
            x_new_index = x_new_index + 1
        else:
            raise ValueError("shouldn't get here")

    return y_new


def get_color_name(idx: int):
    color_names = ['red', 'blue', 'limegreen', 'cyan', 'magenta', 'orange', 'black']
    return color_names[idx % len(color_names)]


def get_line_style(idx: int):
    ls_list = ["-", "--", ":", "-."]
    return ls_list[idx % len(ls_list)]


def get_marker_style(idx: int):
    marker_list = ["o", "v", "*", "^", "s", "P"]
    return marker_list[idx % len(marker_list)]


def xy_plot(img_dir: str,
            df: pd.DataFrame,
            title_1: str,
            title_2: str,
            x_axis_name: str = "Years",
            y_axis_name: str = "",
            expected_df: pd.DataFrame = None,
            fraction_of_total: bool = False,
            show_legend: bool = True,
            show_markers: bool = True,
            min_x: float = None,
            max_x: float = None,
            min_y: float = None,
            max_y: float = None,
            x_axis_as_log_scale: bool = False,
            y_axis_as_log_scale: bool = False):
    """
    Create a plot using the give dataframe, 'df', with all of the appropriate labels.
    The index of the dataframe will be used for the X-axis and the lines or curves will
    be for each column of the dataframe.

    Args:
        img_dir (str, required):
            Directory to save the images. If None, the images will not be saved and a window will be opened.

        df (DataFrame, required):
            A dataframe where the index will be used for the X-values and each column will get
            a separate line/curve.  The name of the column will be used in the legend.

        title_1 (str, required):
            This will be the top line text on the plot.

        title_2 (str, required):
            This will be the second line of text ohe plot.

        x_axis_name (str, optional):
            This is the label used to indicate what the X-axis values are.

        y_axis_name (str, optional):
            This is the label used to indicate what the Y-axis values are.

        expected_df (DataFame, optional):
            This dataframe is expected to have a similar format to the 'df' dataframe.
            The index of the dataframe is the X-values and should similar to that of 'df'.
            The columns are the Y-values and each column creates a separate line/curve.
            However, these will be plotted in black, the markers a little larger, and
            on top of the lines from the 'df'.

        fraction_of_total (bool, optional):
            If true, the columns of each dataframe are summed and divided by this sum
            to create a fraction of the total.

        show_legend (bool, optional):
            If True a legend will be placed on the right side of the plot, but beware
            that long column names can make the plot space very small.

        show_markers (bool, optional):
            if True the lines will have markers at each data point.

        min_x (float, optional):
            If provided the plot will have a fixed minimum value for the X-axis
            independent of the data.  When not provided, matplotlib determines
            the minimum based on the data.

        max_x (float, optional):
            If provided the plot will have a fixed maximum value for the X-axis
            independent of the data.  When not provided, matplotlib determines
            the minimum based on the data.

        min_y (float, optional):
            If provided the plot will have a fixed minimum value for the Y-axis
            independent of the data.  When not provided, matplotlib determines
            the minimum based on the data.

        max_y (float, optional):
            If provided the plot will have a fixed maximum value for the Y-axis
            independent of the data.  When not provided, matplotlib determines
            the minimum based on the data.

        x_axis_as_log_scale (bool, optional):
            If True, the X-axis is assumed to be logrithmic.

        y_axis_as_log_scale (bool, optional):
            If True, the Y-axis is assumed to be logrithmic.

    Returns:
        None - but image will be saved or window opened.
    """

    fig, ax = plt.subplots(layout='constrained')

    if fraction_of_total:
        column_names = df.columns
        df["total"] = 0
        for col_name in column_names:
            df["total"] = df["total"] + df[col_name]
        for col_name in column_names:
            df[col_name] = df[col_name] / df["total"]
        del df["total"]

    x_act_new = df.index.values.tolist()

    color_index = 0
    marker_index = 0
    ls_index = 0

    if expected_df is not None:
        x_exp_old = expected_df.index.values.tolist()
        x_act_old = df.index.values.tolist()
        x_exp_new, x_act_new = create_common_x_values(x_exp_old, x_act_old)

    column_names = df.columns.tolist()
    x_act_old = df.index.values.tolist()

    color_index = 0
    marker_index = 0
    ls_index = 0

    for col_name in column_names:
        y_act_old = df[col_name].values.tolist()

        y_act_new = fill_in_y_values(x_act_old, y_act_old, x_act_new)

        clr = get_color_name(color_index)
        ls  = get_line_style(ls_index)       # noqa: E221
        ms  = get_marker_style(marker_index) # noqa: E221

        if show_markers:
            ax.plot(x_act_new, y_act_new, c=clr, marker=ms, markersize=4, linewidth=1.0, linestyle=ls, label=col_name)
        else:
            ax.plot(x_act_new, y_act_new, c=clr, linewidth=1.0, linestyle=ls, label=col_name)

        marker_index = marker_index + 1
        color_index  = color_index  + 1  # noqa: E221
        ls_index     = ls_index     + 1  # noqa: E221

    if expected_df is not None:
        column_names = expected_df.columns.tolist()

        color_index = 0
        marker_index = 0
        ls_index = 0

        for col_name in column_names:
            y_exp_old = expected_df[col_name].values.tolist()

            y_exp_new = fill_in_y_values(x_exp_old, y_exp_old, x_exp_new)

            clr = get_color_name(6) # 6 = black
            ls  = get_line_style(ls_index)       # noqa: E221
            ms  = get_marker_style(marker_index) # noqa: E221

            if show_markers:
                ax.plot(x_exp_new, y_exp_new, c=clr, marker=ms, markersize=6, linewidth=2.0, linestyle=ls, label=col_name)
            else:
                ax.plot(x_exp_new, y_exp_new, c=clr, linewidth=2.0, linestyle=ls, label=col_name)

            marker_index = marker_index + 1
            ls_index = ls_index + 1
            color_index = color_index + 1

    if x_axis_as_log_scale:
        ax.set_xscale('log')
    elif max_x is not None:
        max = max_x
        min = 0.0
        if min_x is not None:
            min = min_x
        ax.set_xlim([min, max])

    if y_axis_as_log_scale:
        ax.set_yscale('log')
    elif max_y is not None:
        max = max_y
        min = 0.0
        if min_y is not None:
            min = min_y
        ax.set_ylim([min, max])

    ax.minorticks_on()
    ax.grid(which="major", color='darkgray', linewidth=0.5, linestyle='--')
    ax.grid(which="minor", color='gray'    , linewidth=0.1, linestyle='--') # noqa: E203

    if show_legend:
        ax.legend(loc="upper left", bbox_to_anchor=(1.0, 0.75), fontsize='8')

    ax.set_xlabel(x_axis_name)
    ax.set_ylabel(y_axis_name)
    if isinstance(x_act_new[0], str):
        x_act_new = df.index.values.tolist()
        plt.xticks(range(len(x_act_new)), x_act_new, rotation=45, ha='right')
    title = title_1
    if title_2 is not None:
        title = title + "\n" + title_2
    fig.suptitle(title, fontsize='14')

    if img_dir:
        if not os.path.exists(img_dir):
            os.makedirs(img_dir)
        title_1 = title_1.replace(" ", "_")
        fn = os.path.join(img_dir, title_1 + ".png")
        print(fn)
        fig.set_size_inches(9, 6)
        fig.savefig(fn, dpi=300)
    else:
        plt.show()
    plt.close(fig)

    return



from functools import partial
from typing import Dict, Any


class ParameterizedCall:
    def __init__(self,
                 func: callable,
                 non_hyperparameters: Dict[str, Any] = None,
                 hyperparameters: Dict[str, Any] = None,
                 label: str = None):
        self.func = func
        self._non_hyperparameters = {} if non_hyperparameters is None else non_hyperparameters
        self.hyperparameters = {} if hyperparameters is None else hyperparameters
        self.label = label

    @property
    def _label_str(self):
        return "" if self.label is None else f"--{self.label}"

    def _label_hyperparameter(self, hyperparameter: str) -> str:
        labeled_hyperparameter = hyperparameter + self._label_str
        return labeled_hyperparameter

    def _unlabel_hyperparameter(self, labeled_hyperparameter: str) -> str:
        unlabeled_hyperparameter = labeled_hyperparameter.removesuffix(self._label_str)
        if self._label_str != "" and unlabeled_hyperparameter == labeled_hyperparameter:
            raise ValueError(f"Cannot remove label: {self.label} from labeled hyperparameter: {labeled_hyperparameter}")
        return unlabeled_hyperparameter

    @property
    def labeled_hyperparameters(self) -> Dict[str, Any]:
        return {self._label_hyperparameter(hp): value for hp, value in self.hyperparameters.items()}

    @property
    def _hyperparameters_none_filtered(self):
        return {hp: value for hp, value in self.hyperparameters.items() if value is not None}

    def _set_hyperparameter(self, hyperparameter: str, value: Any) -> None:
        if hyperparameter not in self.hyperparameters:
            raise ValueError(f"Cannot set unknown hyperparameter: {hyperparameter}")
        self.hyperparameters[hyperparameter] = value

    def set_labeled_hyperparameter(self, labeled_hyperparameter: str, value: Any) -> None:
        hyperparameter = self._unlabel_hyperparameter(labeled_hyperparameter=labeled_hyperparameter)
        self._set_hyperparameter(hyperparameter=hyperparameter, value=value)

    def prepare_call(self) -> callable:
        return partial(self.func, **self._non_hyperparameters, **self._hyperparameters_none_filtered)



from typing import List, Dict

from emod_api.demographics.updateable import Updateable


# TODO: should we drop NO_GROUP support altogether? If so, group is always INDIVIDUAL_PROPERTY and the instantiation
#  argument should be removed. If NOT, then we need to ensure that an inconsistent state is not possible if calling
#  code toggles "things" e.g. group = 'NO_GROUP' and property_name = 'Risk'
#  https://github.com/InstituteforDiseaseModeling/emodpy-hiv/issues/216

class Assortivity(Updateable):
    def __init__(self,
                 matrix: List[List[float]] = None,
                 group: str = "INDIVIDUAL_PROPERTY"):
        super().__init__()
        self.axes = ['LOW', 'MEDIUM', 'HIGH']
        self.group = group
        self.property_name = 'Risk'
        self.matrix = matrix  # a None matrix is valid if group is 'NO_GROUP'

    def to_dict(self) -> Dict:
        assortivity = {
            'Axes': self.axes,
            'Group': self.group,
            'Property_Name': self.property_name
        }
        if self.matrix is not None:
            assortivity['Weighting_Matrix_RowMale_ColumnFemale'] = self.matrix

        return assortivity

    @classmethod
    def from_dict(cls, d: Dict) -> '__class__':
        keys_to_args = {'Group': 'group', 'Weighting_Matrix_RowMale_ColumnFemale': 'matrix'}
        kwargs = {arg: d[key] for key, arg in keys_to_args.items() if key in d}
        return cls(**kwargs)



import json


def compare_intervention_name(iv_config_1: dict, iv_config_2: dict) -> int:
    """
    Compare two intervention names.
    Returns:
        -1 if iv_config_1 < iv_config_2, 0 if equal, 1 if iv_config_1 > iv_config_2
    """
    iv_name_1 = iv_config_1.get('Intervention_Name', None)
    iv_name_2 = iv_config_2.get('Intervention_Name', None)

    if iv_name_1 is None and iv_name_2 is None:
        return 0
    elif iv_name_1 is None:
        return -1
    elif iv_name_2 is None:
        return 1
    elif iv_name_1 < iv_name_2:
        return -1
    elif iv_name_1 > iv_name_2:
        return 1
    else:
        return 0


def convert_DelayedIntervention(iv_config: dict):
    """
    The old code had HIVDelayedIntervention while the new code uses
    DelayedIntervention with a nested BroadcastEvent class.  This converts
    the DelayedInterventions to be HIVDelayedIntervention.
    """
    if iv_config['class'] == 'DelayedIntervention':
        ac_iv_config_list = iv_config['Actual_IndividualIntervention_Configs']
        if len(ac_iv_config_list) == 1 and ac_iv_config_list[0]['class'] == 'BroadcastEvent':
            iv_config['class'] = 'HIVDelayedIntervention'
            iv_config["Broadcast_Event"] = ac_iv_config_list[0]['Broadcast_Event']
            iv_config.pop("Actual_IndividualIntervention_Configs", None)


def compare_intervention_config(iv_config_1: dict, iv_config_2: dict) -> int:
    """
    Compare two intervention configurations by class name and then intervention name.
    """
    convert_DelayedIntervention(iv_config_1)
    convert_DelayedIntervention(iv_config_2)

    if iv_config_1['class'] < iv_config_2['class']:
        return -1
    elif iv_config_1['class'] > iv_config_2['class']:
        return 1
    else:
        return compare_intervention_name(iv_config_1, iv_config_2)


def compare_sec(event1: dict, event2: dict) -> int:
    """
    Compares the campaign events for two StandardInterventionDistributionEventCoordinators.
    First sort on the intervention configuration, but have special handling if both SECs
    have NLHTIV.
    """
    ec_config_1 = event1['Event_Coordinator_Config']
    ec_config_2 = event2['Event_Coordinator_Config']
    iv_config_1 = ec_config_1['Intervention_Config']
    iv_config_2 = ec_config_2['Intervention_Config']

    if iv_config_1['class'] < iv_config_2['class']:
        return -1
    elif iv_config_1['class'] > iv_config_2['class']:
        return 1
    elif iv_config_1['class'] == 'NodeLevelHealthTriggeredIV':
        tcl_event_1 = iv_config_1['Trigger_Condition_List'][0]
        tcl_event_2 = iv_config_2['Trigger_Condition_List'][0]
        # print(tcl_event_1, tcl_event_2)
        if tcl_event_1 < tcl_event_2:
            # print(tcl_event_1)
            return -1
        elif tcl_event_1 > tcl_event_2:
            # print(tcl_event_2)
            return 1
        else:
            ac_iv_config_1 = iv_config_1['Actual_IndividualIntervention_Config']
            ac_iv_config_2 = iv_config_2['Actual_IndividualIntervention_Config']
            cmp = compare_intervention_config(ac_iv_config_1, ac_iv_config_2)
            if cmp == 0:
                node_set_1 = event1["Nodeset_Config"]
                node_set_2 = event2["Nodeset_Config"]
                return compare_node_set(node_set_1, node_set_2)
            else:
                return cmp
    else:
        cmp = compare_intervention_name(iv_config_1, iv_config_2)
        if cmp == 0:
            node_set_1 = event1["Nodeset_Config"]
            node_set_2 = event2["Nodeset_Config"]
            return compare_node_set(node_set_1, node_set_2)
        else:
            return cmp


def compare_node_set(node_set_1: dict, node_set_2: dict) -> int:
    """
    Sorting by node id is helpful when everything else is similar.
    """
    if node_set_1['class'] < node_set_2['class']:
        return -1
    elif node_set_1['class'] > node_set_2['class']:
        return 1
    elif node_set_1['class'] == 'NodeSetNodeList':
        node_list_1 = node_set_1.get('Node_List', [])
        node_list_2 = node_set_2.get('Node_List', [])
        if len(node_list_1) < len(node_list_2):
            return -1
        elif len(node_list_1) > len(node_list_2):
            return 1
        else:
            for i in range(len(node_list_1)):
                if node_list_1[i] < node_list_2[i]:
                    return -1
                elif node_list_1[i] > node_list_2[i]:
                    return 1
            return 0
    else:
        return 0


def compare_nchooser(event1: dict, event2: dict) -> int:
    """
    Compares the campaign events for two NChooserEventCoordinators.
    First sort on the intervention configuration, then sort on the nodeset configuration.
    Added this logic because the first campaign file for Zambia had an nchooser
    handing out MaleCircumcisions but were different for each node.  This would
    get them in order by node.
    """
    ec_config_1 = event1['Event_Coordinator_Config']
    ec_config_2 = event2['Event_Coordinator_Config']
    iv_config_1 = ec_config_1['Intervention_Config']
    iv_config_2 = ec_config_2['Intervention_Config']

    cmp = compare_intervention_config(iv_config_1, iv_config_2)
    if cmp == 0:
        node_set_1 = event1["Nodeset_Config"]
        node_set_2 = event2["Nodeset_Config"]
        return compare_node_set(node_set_1, node_set_2)
    else:
        return cmp


def compare_tracker(event1: dict, event2: dict) -> int:
    """
    Compares the campaign events for two ReferenceTrackingEventCoordinators.
    First sort on whether the target gender is different and then sort on the
    intervention configuration.
    """
    ec_config_1 = event1['Event_Coordinator_Config']
    ec_config_2 = event2['Event_Coordinator_Config']

    if ec_config_1['Target_Gender'] < ec_config_2['Target_Gender']:
        return -1
    elif ec_config_1['Target_Gender'] > ec_config_2['Target_Gender']:
        return 1
    else:
        iv_config_1 = ec_config_1['Intervention_Config']
        iv_config_2 = ec_config_2['Intervention_Config']
        return compare_intervention_config(iv_config_1, iv_config_2)


def compare_event_coordinator(event1: dict, event2: dict) -> int:
    """
    Compare two campaign events by their coordinator

    Returns:
        -1 if event1 < event2, 0 if equal, 1 if event1 > event2
    """
    ec_config_1 = event1['Event_Coordinator_Config']
    ec_config_2 = event2['Event_Coordinator_Config']

    if ec_config_1['class'] < ec_config_2['class']:
        return -1
    elif ec_config_1['class'] > ec_config_2['class']:
        return 1
    elif ec_config_1['class'] == 'StandardInterventionDistributionEventCoordinator':
        return compare_sec(event1, event2)
    elif ec_config_1['class'].startswith('NChooserEventCoordinator'):
        return compare_nchooser(event1, event2)
    elif ec_config_1['class'].startswith('ReferenceTrackingEventCoordinator'):
        return compare_tracker(event1, event2)
    else:
        return 0


def compare_campaign_event(event1: dict, event2: dict) -> int:
    """
    Compare two campaign events

    Returns:
        -1 if event1 < event2, 0 if equal, 1 if event1 > event2
    """
    if event1['class'] < event2['class']:
        return -1
    elif event1['class'] > event2['class']:
        return 1
    elif event1['class'] == 'CampaignEventByYear':
        if event1['Start_Year'] < event2['Start_Year']:
            return -1
        elif event1['Start_Year'] > event2['Start_Year']:
            return 1
        else:
            return compare_event_coordinator(event1, event2)
    elif event1['class'] == 'CampaignEvent':
        start_day_1 = event1.get('Start_Day', None)
        start_day_2 = event2.get('Start_Day', None)
        if start_day_1 is None and start_day_2 is None:
            return compare_event_coordinator(event1, event2)
        elif start_day_1 is None:
            return -1
        elif start_day_2 is None:
            return 1
        elif start_day_1 < start_day_2:
            return -1
        elif start_day_1 > start_day_2:
            return 1
        else:
            return compare_event_coordinator(event1, event2)


def sort_names_and_probabilities(iv_config: dict):
    """
    Sorting the names and probabilities makes sure that two campaign files
    have them in the same order.  If the numbers are exactly the same,
    being in the same order will make it obvious.
    """
    if iv_config['class'] == 'HIVRandomChoice':
        names = iv_config.get("Choice_Names", [])
        probs = iv_config.get("Choice_Probabilities", [])
        for i in range(len(probs) - 1):
            for j in range(i + 1, len(probs)):
                if probs[i] < probs[j]:
                    # Swap probabilities
                    probs[i], probs[j] = probs[j], probs[i]
                    # Swap names
                    names[i], names[j] = names[j], names[i]


def remove_defaults_from_intervention(iv_config: dict):
    """
    Remove default values from the intervention configuration.
    """

    # common parameters and their defaults
    if 'Dont_Allow_Duplicates' in iv_config and iv_config['Dont_Allow_Duplicates'] == 0:
        iv_config.pop("Dont_Allow_Duplicates", None)
    if 'New_Property_Value' in iv_config and iv_config['New_Property_Value'] == "":
        iv_config.pop("New_Property_Value", None)
    if 'Intervention_Name' in iv_config and iv_config['Intervention_Name'] == iv_config['class']:
        iv_config.pop("Intervention_Name", None)
    if 'Disqualifying_Properties' in iv_config and len(iv_config['Disqualifying_Properties']) == 0:
        iv_config.pop("Disqualifying_Properties", None)
    if 'Event_Or_Config' in iv_config:
        iv_config.pop("Event_Or_Config", None)

    # consider the specific intervention classes
    if iv_config['class'] == 'DelayedIntervention':
        if 'Coverage' in iv_config and iv_config['Coverage'] == 1:
            iv_config.pop("Coverage", None)
        if 'Delay_Period_Min' in iv_config and iv_config['Delay_Period_Min'] == 0:
            iv_config.pop("Delay_Period_Min", None)
        if (('Intervention_Name' in iv_config) and ((iv_config['Intervention_Name'] == 'DelayedIntervention') or (iv_config['Intervention_Name'] == 'HIVDelayedIntervention'))):
            iv_config.pop("Intervention_Name", None)
        ac_iv_config_list = iv_config['Actual_IndividualIntervention_Configs']
        for config in ac_iv_config_list:
            remove_defaults_from_intervention(config)

    elif iv_config['class'] == 'HIVDelayedIntervention':
        if 'Coverage' in iv_config and iv_config['Coverage'] == 1:
            iv_config.pop("Coverage", None)
        if 'Delay_Period_Min' in iv_config and iv_config['Delay_Period_Min'] == 0:
            iv_config.pop("Delay_Period_Min", None)
        if 'Delay_Period_Min' in iv_config and iv_config['Delay_Period_Min'] == 0:
            iv_config.pop("Delay_Period_Min", None)
        if (('Intervention_Name' in iv_config) and ((iv_config['Intervention_Name'] == 'DelayedIntervention') or (iv_config['Intervention_Name'] == 'HIVDelayedIntervention'))):
            iv_config.pop("Intervention_Name", None)

    elif iv_config['class'] == 'HIVRapidHIVDiagnostic':
        if 'Base_Sensitivity' in iv_config and iv_config['Base_Sensitivity'] == 1:
            iv_config.pop("Base_Sensitivity", None)
        if 'Base_Specificity' in iv_config and iv_config['Base_Specificity'] == 1:
            iv_config.pop("Base_Specificity", None)
        if 'Cost_To_Consumer' in iv_config and iv_config['Cost_To_Consumer'] == 1:
            iv_config.pop("Cost_To_Consumer", None)
        if 'Days_To_Diagnosis' in iv_config and iv_config['Days_To_Diagnosis'] == 0:
            iv_config.pop("Days_To_Diagnosis", None)
        if 'Enable_Is_Symptomatic' in iv_config and iv_config['Enable_Is_Symptomatic'] == 0:
            iv_config.pop("Enable_Is_Symptomatic", None)
        if 'Negative_Diagnosis_Event' in iv_config and iv_config['Negative_Diagnosis_Event'] == "":
            iv_config.pop("Negative_Diagnosis_Event", None)
        if 'Probability_Received_Result' in iv_config and iv_config['Probability_Received_Result'] == 1:
            iv_config.pop("Probability_Received_Result", None)
        if 'Sensitivity_Type' in iv_config and iv_config['Sensitivity_Type'] == "SINGLE_VALUE":
            iv_config.pop("Sensitivity_Type", None)

    elif iv_config['class'] == 'HIVSigmoidByYearAndSexDiagnostic':
        # "Female_Multiplier": 1,
        if 'Female_Multiplier' in iv_config and iv_config['Female_Multiplier'] == 1:
            iv_config.pop("Female_Multiplier", None)
        if 'Ramp_Min' in iv_config and iv_config['Ramp_Min'] == 0:
            iv_config.pop("Ramp_Min", None)
        if 'Ramp_Max' in iv_config and iv_config['Ramp_Max'] == 1:
            iv_config.pop("Ramp_Max", None)
        if 'Ramp_Rate' in iv_config and iv_config['Ramp_Rate'] == 1:
            iv_config.pop("Ramp_Rate", None)

    elif iv_config['class'] == 'HIVMuxer':
        if 'Broadcast_On_Expiration_Event' in iv_config and iv_config['Broadcast_On_Expiration_Event'] == "":
            iv_config.pop("Broadcast_On_Expiration_Event", None)
        if 'Coverage' in iv_config and iv_config['Coverage'] == 1:
            iv_config.pop("Coverage", None)
        if 'Expiration_Period' in iv_config and iv_config['Expiration_Period'] == 3.40282e+38:
            iv_config.pop("Expiration_Period", None)
        if 'Max_Entries' in iv_config and iv_config['Max_Entries'] == 1:
            iv_config.pop("Max_Entries", None)

    elif ((iv_config['class'] == 'HIVARTStagingCD4AgnosticDiagnostic') or (iv_config['class'] == 'HIVARTStagingByCD4Diagnostic')):
        if 'Adult_Treatment_Age' in iv_config and iv_config['Adult_Treatment_Age'] == 5:
            iv_config.pop("Adult_Treatment_Age", None)
        if 'Individual_Property_Active_TB_Key' in iv_config and iv_config['Individual_Property_Active_TB_Key'] == "UNINITIALIZED":
            iv_config.pop("Individual_Property_Active_TB_Key", None)
        if 'Individual_Property_Active_TB_Value' in iv_config and iv_config['Individual_Property_Active_TB_Value'] == "UNINITIALIZED":
            iv_config.pop("Individual_Property_Active_TB_Value", None)

    elif iv_config['class'] == 'HIVPiecewiseByYearAndSexDiagnostic':
        if 'Default_Value' in iv_config and iv_config['Default_Value'] == 0:
            iv_config.pop("Default_Value", None)
        if 'Interpolation_Order' in iv_config and iv_config['Interpolation_Order'] == 0:
            iv_config.pop("Interpolation_Order", None)
        if 'Female_Multiplier' in iv_config and iv_config['Female_Multiplier'] == 1:
            iv_config.pop("Female_Multiplier", None)
        if 'Negative_Diagnosis_Event' in iv_config and iv_config['Negative_Diagnosis_Event'] == "":
            iv_config.pop("Negative_Diagnosis_Event", None)

    elif iv_config['class'] == 'PropertyValueChanger':
        if 'Daily_Probability' in iv_config and iv_config['Daily_Probability'] == 1:
            iv_config.pop("Daily_Probability", None)
        if 'Maximum_Duration' in iv_config and iv_config['Maximum_Duration'] == 3.40282e+38:
            iv_config.pop("Maximum_Duration", None)
        if 'Revert' in iv_config and iv_config['Revert'] == 0:
            iv_config.pop("Revert", None)

    elif iv_config['class'] == 'MaleCircumcision':
        if 'Apply_If_Higher_Reduced_Acquire' in iv_config and iv_config['Apply_If_Higher_Reduced_Acquire'] == 0:
            iv_config.pop("Apply_If_Higher_Reduced_Acquire", None)
        if 'Circumcision_Reduced_Acquire' in iv_config and iv_config['Circumcision_Reduced_Acquire'] == 0.6:
            iv_config.pop("Circumcision_Reduced_Acquire", None)

    elif iv_config['class'] == 'HIVRandomChoice':
        # sorting makes these easier to compare
        sort_names_and_probabilities(iv_config)

    elif iv_config['class'] == 'OutbreakIndividual':
        if 'Antigen' in iv_config and iv_config['Antigen'] == 0:
            iv_config.pop("Antigen", None)
        if 'Genome' in iv_config and iv_config['Genome'] == 0:
            iv_config.pop("Genome", None)
        if 'Ignore_Immunity' in iv_config and iv_config['Ignore_Immunity'] == 1:
            iv_config.pop("Ignore_Immunity", None)
        if 'Outbreak_Source' in iv_config:
            iv_config.pop("Outbreak_Source", None)
        if 'Event_Name' in iv_config:
            iv_config.pop("Event_Name", None)


def remove_defaults_from_xxx_target_demographic(config: dict):
    """
    Remove the defaults that are common to SEC and NLHTIV
    """
    if 'Property_Restrictions' in config and len(config['Property_Restrictions']) == 0:
        config.pop("Property_Restrictions", None)
    if 'Property_Restrictions_Within_Node' in config and len(config['Property_Restrictions_Within_Node']) == 0:
        config.pop("Property_Restrictions_Within_Node", None)
    if 'Targeting_Config' in config and len(config['Targeting_Config']) == 0:
        config.pop("Targeting_Config", None)
    if 'Demographic_Coverage' in config and config['Demographic_Coverage'] == 1:
        config.pop("Demographic_Coverage", None)
    if 'Target_Residents_Only' in config and config['Target_Residents_Only'] == 0:
        config.pop("Target_Residents_Only", None)
    if 'Target_Demographic' in config and config['Target_Demographic'] == "Everyone":
        config.pop("Target_Demographic", None)
    if 'Target_Gender' in config and config['Target_Gender'] == "All":
        config.pop("Target_Gender", None)
    if 'Travel_Linked' in config:
        config.pop("Travel_Linked", None)


def remove_defaults_from_sec(ec_config: dict):
    """
    Remove default values from the StandardInterventionDistributionEventCoordinator configuration
    and its associated intervention configurations
    """

    # SEC specific defaults
    if 'Node_Property_Restrictions' in ec_config and len(ec_config['Node_Property_Restrictions']) == 0:
        ec_config.pop("Node_Property_Restrictions", None)
    if 'Number_Repetitions' in ec_config and ec_config['Number_Repetitions'] == 1:
        ec_config.pop("Number_Repetitions", None)
    if 'Timesteps_Between_Repetitions' in ec_config and ec_config['Timesteps_Between_Repetitions'] == -1:
        ec_config.pop("Timesteps_Between_Repetitions", None)
    if 'Individual_Selection_Type' in ec_config and ec_config['Individual_Selection_Type'] == "DEMOGRAPHIC_COVERAGE":
        ec_config.pop("Individual_Selection_Type", None)

    # defaults common with NLHTIV
    remove_defaults_from_xxx_target_demographic(ec_config)

    iv_config = ec_config.get('Intervention_Config', {})
    remove_defaults_from_intervention(iv_config)

    if iv_config.get('class', None) == 'NodeLevelHealthTriggeredIV':
        remove_defaults_from_xxx_target_demographic(iv_config)

        # NLHTIV specific defaults
        if 'Blackout_Event_Trigger' in iv_config and iv_config['Blackout_Event_Trigger'] == "":
            iv_config.pop("Blackout_Event_Trigger", None)
        if 'Blackout_On_First_Occurrence' in iv_config and iv_config['Blackout_On_First_Occurrence'] == 0:
            iv_config.pop("Blackout_On_First_Occurrence", None)
        if 'Blackout_Period' in iv_config and iv_config['Blackout_Period'] == 0:
            iv_config.pop("Blackout_Period", None)
        if 'Distribute_On_Return_Home' in iv_config and iv_config['Distribute_On_Return_Home'] == 0:
            iv_config.pop("Distribute_On_Return_Home", None)
        if 'Duration' in iv_config and iv_config['Duration'] == -1:
            iv_config.pop("Duration", None)
        if 'Node_Property_Restrictions' in iv_config and len(iv_config['Node_Property_Restrictions']) == 0:
            iv_config.pop("Node_Property_Restrictions", None)

        ac_iv_config = iv_config['Actual_IndividualIntervention_Config']
        remove_defaults_from_intervention(ac_iv_config)


def fix_nchooser_number_format(ec_config: dict):
    """
    The Nchoosers used for distributing VMMC had lots of places where it was comparing
    15 vs 15.0 or 14.999 vs 14.999999.  This made comparisons difficult.
    """
    distributions = ec_config.get("Distributions", [])
    for dist in distributions:
        for age_range in dist.get("Age_Ranges_Years", []):
            age_range["Max"] = int(age_range["Max"]) + 0.999
            age_range["Min"] = int(age_range["Min"])
        dist["End_Year"] = int(dist["End_Year"]) + 0.999
        num_targeted = ec_config.get("Num_Targeted", [])
        if len(num_targeted) == 0:
            dist.pop("Num_Targeted", None)


def remove_defaults(campaign_json: dict):
    """
    Remove default values from the campaign JSON.  This makes the resulting file smaller,
    easier to read, and easier to compare with other campaign files.
    """
    for event in campaign_json.get("Events", []):
        ec_config = event.get("Event_Coordinator_Config", {})
        if ec_config['class'] == 'StandardInterventionDistributionEventCoordinator':
            remove_defaults_from_sec(ec_config)
        elif ec_config['class'].startswith('NChooserEventCoordinator'):
            fix_nchooser_number_format(ec_config)
            iv_config = ec_config["Intervention_Config"]
            remove_defaults_from_intervention(iv_config)


def sort_campaign(campaign_json: dict):
    """
    Sorts the campaign events in the given campaign JSON.
    The sorting is done based on the event class, start year, and event coordinator configuration."""

    # -------------------------------------------------------------------------------------
    # --- NOTE:  I'm using my own bubble sort implementation here because I wasn't getting
    # --- what I wanted with the built-in sorted() function.
    # -------------------------------------------------------------------------------------

    n = len(campaign_json["Events"])
    for i in range(n - 1):
        for j in range(i + 1, n):
            # print(f"Comparing event {i} with event {j}")
            event_i = campaign_json["Events"][i]
            event_j = campaign_json["Events"][j]
            cmp = compare_campaign_event(event_i, event_j)
            if cmp > 0:
                campaign_json["Events"][i], campaign_json["Events"][j] = campaign_json["Events"][j], campaign_json["Events"][i]

    return campaign_json


def sort_campaign_file(existing_filename: str, new_filename: str = None):
    """
    Sorts the campaign events in a given JSON file and writes the sorted events back to a file.
    """
    with open(existing_filename, 'r') as f:
        campaign_json = json.load(f)

    campaign_json = sort_campaign(campaign_json)

    # removing the defaults makes the file smaller, easier to read, and easier to compare
    # with other campaign files.
    remove_defaults(campaign_json)

    with open(new_filename, 'w') as f:
        campaign_json = json.dumps(campaign_json, indent=4, sort_keys=True)
        campaign_json = json.loads(campaign_json, parse_float=lambda x: round(float(x), 9))
        json.dump(campaign_json, f, indent=4, sort_keys=True)


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()

    parser.add_argument('existing_filename', type=str, nargs=1, help='Name of campaign file to sort.')
    parser.add_argument('new_filename', type=str, default=None, nargs='?',
                        help='Name of new campaign file to write sorted events to. '
                        + 'If not provided, will overwrite the existing file.')

    args = parser.parse_args()

    existing_filename = args.existing_filename[0]
    new_filename = args.new_filename if args.new_filename else args.existing_filename[0]

    sort_campaign_file(existing_filename=existing_filename, new_filename=new_filename)
    print(f"Sorted campaign events from {existing_filename} and saved to {new_filename}.")



from importlib import resources
from pathlib import Path
from typing import List, Union
import copy

import pandas as pd

import emod_api
from emod_api.schema_to_class import ReadOnlyDict

import emodpy_hiv.countries.zambia as zambia_data
from emodpy_hiv.country_model import Country

import emodpy_hiv.demographics.infer_natural_mortality as infer

from emodpy_hiv.demographics.hiv_demographics import HIVDemographics
from emodpy_hiv.demographics.relationship_types import RelationshipTypes
from emodpy_hiv.demographics.risk_groups import RiskGroups
from emodpy_hiv.parameterized_call import ParameterizedCall
from emodpy_hiv.reporters.reporters import (Reporters, ReportFilter, ReportHIVInfection, ReportHIVART,
                                            ReportHIVMortality, ReportHIVByAgeAndGender,
                                            ReportTransmission, ReportRelationshipStart, InsetChart)

COMMERCIAL = RelationshipTypes.commercial.value
TRANSITORY = RelationshipTypes.transitory.value
INFORMAL   = RelationshipTypes.informal.value   # noqa: E221
MARITAL    = RelationshipTypes.marital.value    # noqa: E221

LOW    = RiskGroups.low.value    # noqa: E221
MEDIUM = RiskGroups.medium.value
HIGH   = RiskGroups.high.value   # noqa: E221

# The nodes for the country with names and initial population.
# NOTE: In the future, please add information about there this initial population data comes from.
NODE_DATA = [
    {"node_id":  1, "name": "Central",      "population": 304498},  # noqa: E241
    {"node_id":  2, "name": "Copperbelt",   "population": 459461},  # noqa: E241
    {"node_id":  3, "name": "Eastern",      "population": 371018},  # noqa: E241
    {"node_id":  4, "name": "Luapula",      "population": 231074},  # noqa: E241
    {"node_id":  5, "name": "Lusaka",       "population": 510456},  # noqa: E241
    {"node_id":  6, "name": "Muchinga",     "population": 165784},  # noqa: E241
    {"node_id":  7, "name": "Northwestern", "population": 169368},  # noqa: E241
    {"node_id":  8, "name": "Northern",     "population": 257607},  # noqa: E241
    {"node_id":  9, "name": "Southern",     "population": 370381},  # noqa: E241
    {"node_id": 10, "name": "Western",      "population": 210352}   # noqa: E241
]

# Convert the data above so that we can use the country name
# when creating labels for parameters.
NODE_ID_TO_NODE_NAME = {node_dict["node_id"]: node_dict["name"] for node_dict in NODE_DATA}


class Zambia(Country):
    """
    Zambia country model class. This class is a subclass of the Country class and implements the build_campaign method
    to build campaign object to be used with EMODTask.

    Example:
        >>> from emodpy_hiv.countries.zambia import Zambia
        >>> from emodpy.emod_task import EMODTask
        >>> import manifest
        >>> from functools import partial
        >>> # Define a campaign_builder function to pass in the required arguments for the Zambia.build_campaign() method.
        >>> def campaign_builder():
        >>>     return Zambia.build_campaign(schema_path=manifest.schema_file, base_year=Zambia.base_year)
        >>> # Alternatively, you can use the partial function to pass in the required arguments for the
        >>> # Zambia.build_campaign() method. For example:
        >>> # campaign_builder = partial(Zambia.build_campaign, schema_path=manifest.schema_file, base_year=Zambia.base_year)
        >>> # Create an EMODTask object using the campaign_builder function
        >>> task = EMODTask.from_defaults(config_path="config.json",
        >>>                               eradication_path=manifest.eradication_path,
        >>>                               campaign_builder=campaign_builder, ...)
    """

    country_name = "Zambia"
    _historical_vmmc_data_file = resources.files(zambia_data).joinpath("historical_vmmc_data.csv")
    _inital_demog_cache = None

    @classmethod
    def initialize_config(cls, schema_path: Union[str, Path]) -> ReadOnlyDict:
        config = super().initialize_config(schema_path=schema_path)

        config.parameters.Simulation_Type = "HIV_SIM"
        config.parameters.Simulation_Duration = 99.5 * 365
        config.parameters.Simulation_Timestep = 30.4166666666667

        # This is the current "production scale" population scaling factor for Zambia the model is calibrated to
        config.parameters.x_Base_Population = 0.05
        # x_Base_Population = 0.05 needs more than 8GB memory, so we're giving it here
        config.parameters.Memory_Usage_Halting_Threshold_Working_Set_MB = 12000
        config.parameters.Memory_Usage_Warning_Threshold_Working_Set_MB = 11000

        config.parameters.Start_Time = 0
        config.parameters.AIDS_Duration_In_Months = 9
        config.parameters.AIDS_Stage_Infectivity_Multiplier = 4.5
        config.parameters.Acute_Duration_In_Months = 2.9
        config.parameters.Acute_Stage_Infectivity_Multiplier = 26
        config.parameters.Age_Initialization_Distribution_Type = "DISTRIBUTION_COMPLEX"
        config.parameters.Base_Individual_Sample_Rate = 1
        config.parameters.Base_Infectivity = 0.00065805
        config.parameters.Base_Year = 1960.5
        config.parameters.Birth_Rate_Dependence = "INDIVIDUAL_PREGNANCIES_BY_AGE_AND_YEAR"
        config.parameters.Birth_Rate_Time_Dependence = "NONE"
        config.parameters.CD4_At_Death_LogLogistic_Heterogeneity = 0.7
        config.parameters.CD4_At_Death_LogLogistic_Scale = 2.96
        config.parameters.CD4_Post_Infection_Weibull_Heterogeneity = 0.2756
        config.parameters.CD4_Post_Infection_Weibull_Scale = 560.43
        config.parameters.Coital_Dilution_Factor_2_Partners = 0.75
        config.parameters.Coital_Dilution_Factor_3_Partners = 0.6
        config.parameters.Coital_Dilution_Factor_4_Plus_Partners = 0.45
        config.parameters.Condom_Transmission_Blocking_Probability = 0.8
        config.parameters.Days_Between_Symptomatic_And_Death_Weibull_Heterogeneity = 0.5
        config.parameters.Days_Between_Symptomatic_And_Death_Weibull_Scale = 618.341625
        config.parameters.Death_Rate_Dependence = "NONDISEASE_MORTALITY_BY_YEAR_AND_AGE_FOR_EACH_GENDER"
        config.parameters.Enable_Aging = 1
        config.parameters.Enable_Birth = 1
        config.parameters.Enable_Coital_Dilution = 1
        config.parameters.Enable_Demographics_Reporting = 0  # Demographics reporting is on by default, turning it off
        config.parameters.Enable_Infectivity_Reservoir = 0
        config.parameters.Enable_Maternal_Protection = 0
        config.parameters.Enable_Natural_Mortality = 1
        config.parameters.HIV_Adult_Survival_Scale_Parameter_Intercept = 21.182
        config.parameters.HIV_Adult_Survival_Scale_Parameter_Slope = -0.2717
        config.parameters.HIV_Adult_Survival_Shape_Parameter = 2
        config.parameters.HIV_Age_Max_for_Adult_Age_Dependent_Survival = 50
        config.parameters.HIV_Age_Max_for_Child_Survival_Function = 15
        config.parameters.HIV_Child_Survival_Rapid_Progressor_Fraction = 0.57
        config.parameters.HIV_Child_Survival_Rapid_Progressor_Rate = 1.52
        config.parameters.HIV_Child_Survival_Slow_Progressor_Scale = 16
        config.parameters.HIV_Child_Survival_Slow_Progressor_Shape = 2.7
        config.parameters.Heterogeneous_Infectiousness_LogNormal_Scale = 0
        config.parameters.Incubation_Period_Constant = 0
        config.parameters.Incubation_Period_Distribution = "CONSTANT_DISTRIBUTION"
        config.parameters.Individual_Sampling_Type = "FIXED_SAMPLING"
        config.parameters.Infection_Updates_Per_Timestep = 1
        config.parameters.Infectivity_Scale_Type = "CONSTANT_INFECTIVITY"
        config.parameters.Male_To_Female_Relative_Infectivity_Ages = [0, 15, 25]
        config.parameters.Male_To_Female_Relative_Infectivity_Multipliers = [
            2.26154624,
            2.26154624,
            1.112477558
        ]
        config.parameters.Maternal_Infection_Transmission_Probability = 0.3
        config.parameters.Maternal_Transmission_ART_Multiplier = 0.03334
        config.parameters.Migration_Model = "NO_MIGRATION"
        config.parameters.Min_Days_Between_Adding_Relationships = 0
        config.parameters.Node_Grid_Size = 0.009
        config.parameters.PFA_Burnin_Duration_In_Days = 5475
        config.parameters.PFA_Cum_Prob_Selection_Threshold = 0.2
        config.parameters.Population_Density_Infectivity_Correction = "CONSTANT_INFECTIVITY"
        config.parameters.Population_Scale_Type = "FIXED_SCALING"
        config.parameters.Run_Number = 1
        config.parameters.STI_Coinfection_Acquisition_Multiplier = 5.5
        config.parameters.STI_Coinfection_Transmission_Multiplier = 5.5
        config.parameters.Serialized_Population_Reading_Type = "NONE"
        config.parameters.Serialized_Population_Writing_Type = "NONE"
        config.parameters.Sexual_Debut_Age_Setting_Type = "WEIBULL"
        config.parameters.Sexual_Debut_Age_Female_Weibull_Heterogeneity = 0.22108015674142
        config.parameters.Sexual_Debut_Age_Female_Weibull_Scale = 19.9609856656853
        config.parameters.Sexual_Debut_Age_Male_Weibull_Heterogeneity = 0.165108885953777
        config.parameters.Sexual_Debut_Age_Male_Weibull_Scale = 16.8039276633964
        config.parameters.Sexual_Debut_Age_Min = 13

        config.parameters["logLevel_InfectionHIV"] = "ERROR"
        config.parameters["logLevel_Instrumentation"] = "INFO"
        config.parameters["logLevel_Memory"] = "INFO"
        config.parameters["logLevel_OutbreakIndividual"] = "ERROR"
        config.parameters["logLevel_Simulation"] = "INFO"
        config.parameters["logLevel_SusceptibilityHIV"] = "ERROR"
        config.parameters["logLevel_AntiretroviralTherapy"] = "ERROR"
        config.parameters["logLevel_default"] = "WARNING"

        return config

    @classmethod
    def initialize_campaign(cls, schema_path: Union[str, Path]) -> emod_api.campaign:
        campaign = super().initialize_campaign(schema_path=schema_path)
        return campaign

    #
    # Campaign building below
    #

    @classmethod
    def add_male_circumcision_parameterized_calls(cls) -> List[ParameterizedCall]:
        parameterized_calls = []

        # ---- Add historical vmmc NChooser events to campaign ----
        non_hp = {'historical_vmmc_data_filepath': cls._historical_vmmc_data_file}
        hp = {'historical_vmmc_reduced_acquire': None}
        pc = ParameterizedCall(func=cls.add_historical_vmmc_nchooser, non_hyperparameters=non_hp, hyperparameters=hp)
        parameterized_calls.append(pc)

        # ---- Add traditional male circumcision to campaign ----
        traditional_male_circumcision_coverage_by_node = {
            1: 0.054978651,
            2: 0.139462861,
            3: 0.028676043,
            4: 0.091349358,
            5: 0.12318707,
            6: 0.039308099,
            7: 0.727917322,
            8: 0.041105263,
            9: 0.044388102,
            10: 0.398239794
        }

        for node_id, coverage in traditional_male_circumcision_coverage_by_node.items():
            hp = {
                'traditional_male_circumcision_coverage': coverage,
                'traditional_male_circumcision_reduced_acquire': None
            }
            pc = ParameterizedCall(func=cls.add_traditional_male_circumcision,
                                   non_hyperparameters={'node_ids': [node_id]},
                                   hyperparameters=hp,
                                   label=NODE_ID_TO_NODE_NAME[node_id])
            parameterized_calls.append(pc)

        # ---- Add (VMMC) reference tracking to campaign ----
        pc = ParameterizedCall(func=cls.add_vmmc_reference_tracking,
                               hyperparameters={'tracking_vmmc_reduced_acquire': None})
        parameterized_calls.append(pc)

        return parameterized_calls

    @classmethod
    def add_pmtct_parameterized_calls(cls) -> List[ParameterizedCall]:
        parameterized_calls = []

        pc = ParameterizedCall(func=cls.add_pmtct)
        parameterized_calls.append(pc)

        return parameterized_calls

    @classmethod
    def add_seed_infections_parameterized_calls(cls) -> List[ParameterizedCall]:
        seeding_data = []
        seeding_data.append({"node_id":  1, "start_year": 1977.96490840667, "coverage": 0.00994069137813979}) # noqa: E241
        seeding_data.append({"node_id":  2, "start_year": 1981.21526571588, "coverage": 0.0290879518815339}) # noqa: E241, E202
        seeding_data.append({"node_id":  3, "start_year": 1987.89402149164, "coverage": 0.11194470109949}) # noqa: E241
        seeding_data.append({"node_id":  4, "start_year": 1970.14036061384, "coverage": 0.0222773645719442}) # noqa: E241
        seeding_data.append({"node_id":  5, "start_year": 1970.36530180447, "coverage": 0.35030772723087}) # noqa: E241, E202
        seeding_data.append({"node_id":  6, "start_year": 1984.22541975358, "coverage": 0.0167949011180618}) # noqa: E241, E202
        seeding_data.append({"node_id":  7, "start_year": 1980.54040493867, "coverage": 0.0930776686478119}) # noqa: E241
        seeding_data.append({"node_id":  8, "start_year": 1979.51682308393, "coverage": 0.0099968647453273}) # noqa: E241
        seeding_data.append({"node_id":  9, "start_year": 1984.36462983067, "coverage": 0.0731563905879651}) # noqa: E241
        seeding_data.append({"node_id": 10, "start_year": 1982.02918538299, "coverage": 0.267106173241476})

        parameterized_calls = []
        for data in seeding_data:
            hp = {'seeding_start_year': data["start_year"], 'seeding_coverage': data["coverage"]}
            nhp = {'node_ids': [data["node_id"]]}
            pc = ParameterizedCall(func=cls.seed_infections,
                                   non_hyperparameters=nhp,
                                   hyperparameters=hp,
                                   label=NODE_ID_TO_NODE_NAME[data["node_id"]])
            parameterized_calls.append(pc)

        return parameterized_calls

    @classmethod
    def add_csw_parameterized_calls(cls) -> List[ParameterizedCall]:
        parameterized_calls = []

        pc = ParameterizedCall(func=cls.add_csw,
                               non_hyperparameters={'node_ids': [1, 2, 3, 7, 9]},
                               hyperparameters={'csw_male_uptake_coverage': 0.195, 'csw_female_uptake_coverage': 0.125},
                               label='nodeset1')
        parameterized_calls.append(pc)

        pc = ParameterizedCall(func=cls.add_csw,
                               non_hyperparameters={'node_ids': [4, 5, 6, 8, 10]},
                               hyperparameters={'csw_male_uptake_coverage': 0.0781, 'csw_female_uptake_coverage': 0.05},
                               label='nodeset2')
        parameterized_calls.append(pc)

        return parameterized_calls

    @classmethod
    def add_post_debut_coinfection_parameterized_calls(cls) -> List[ParameterizedCall]:
        parameterized_calls = []

        hp = {'coinfection_coverage_HIGH': None, 'coinfection_coverage_MEDIUM': None, 'coinfection_coverage_LOW': None}
        pc = ParameterizedCall(func=cls.add_post_debut_coinfection, hyperparameters=hp)
        parameterized_calls.append(pc)

        return parameterized_calls

    @classmethod
    def add_health_care_testing_parameterized_calls(cls) -> List[ParameterizedCall]:
        parameterized_calls = []
        non_hyperparameters = {
            'hct_delay_to_next_test': [730, 365, 1100],
            'hct_delay_to_next_test_node_ids': [[1, 2, 3, 4, 6, 7], [5, 9, 10], [8]],
            'hct_delay_to_next_test_node_names': ['Default', 'Lusaka, Southern, Western', 'Northern']
        }
        pc = ParameterizedCall(func=cls.add_health_care_testing,
                               non_hyperparameters=non_hyperparameters,
                               hyperparameters={'hct_start_year': None})
        parameterized_calls.append(pc)

        return parameterized_calls

    @classmethod
    def add_ART_cascade_parameterized_calls(cls) -> List[ParameterizedCall]:
        parameterized_calls = []

        hp = {
            'art_cascade_start_year': None,
            'cd4_retention_rate': None,
            'linking_to_pre_art_sigmoid_min': 0.824382156,
            'linking_to_pre_art_sigmoid_max': 0.959334475,
            'linking_to_pre_art_sigmoid_midyear': 2008.506768,
            'linking_to_pre_art_sigmoid_rate': None,
            'pre_staging_retention': None,
            'pre_art_retention': None,
            'linking_to_art_sigmoid_min': None,
            'linking_to_art_sigmoid_max': 0.90041509,
            'linking_to_art_sigmoid_midyear': 2001.053678,
            'linking_to_art_sigmoid_rate': None,
            'art_reenrollment_willingness': None,
            'immediate_art_rate': None
        }
        pc = ParameterizedCall(func=cls.add_ART_cascade, hyperparameters=hp)
        parameterized_calls.append(pc)

        return parameterized_calls

    @classmethod
    def get_campaign_parameterized_calls(cls, campaign: emod_api.campaign) -> List[ParameterizedCall]:
        parameterized_calls = super().get_campaign_parameterized_calls(campaign=campaign)

        # ---- Add male circumcision to campaign -----
        # This block contains 3 chunk of events: historical vmmc NChooser, traditional male circumcision and VMMC with
        # reference tracking.
        parameterized_calls.extend(cls.add_male_circumcision_parameterized_calls())
        # ---- Add prevention of mother-to-child transmission (PMTCT) to campaign ----
        # This pmtct block contains 2 Cascade states: TestingOnANC and TestingOnChild6w
        parameterized_calls.extend(cls.add_pmtct_parameterized_calls())
        # ---- Add OutbreakIndividual interventions to campaign ----
        parameterized_calls.extend(cls.add_seed_infections_parameterized_calls())
        # ---- Add commercial sex worker (CSW) uptake and dropout (with delays) for men and women to campaign ----
        parameterized_calls.extend(cls.add_csw_parameterized_calls())
        # ---- Add co-infections post sexual debut to campaign ----
        parameterized_calls.extend(cls.add_post_debut_coinfection_parameterized_calls())
        # ---- Add health care testing loops to campaign ----
        # This health care testing block contains 3 Cascade states: HCTUptakeAtDebut, HCTUptakePostDebut and
        # HCTTestingLoop
        parameterized_calls.extend(cls.add_health_care_testing_parameterized_calls())
        # ---- Add Antiretroviral Therapy (ART) Cascade to campaign ----
        # This ART Cascade block contains 8 Cascade states: TestingOnSymptomatic, ARTStagingDiagnosticTest, ARTStaging
        # LinkingToPreART, OnPreART, LinkingToART, OnART and LostForever
        parameterized_calls.extend(cls.add_ART_cascade_parameterized_calls())

        return parameterized_calls

    #
    # Demographics building below
    #

    @classmethod
    def initialize_demographics(cls) -> HIVDemographics:
        import emodpy_hiv.demographics.un_world_pop as unwp

        # ------------------------------------------------------------------------------------
        # --- The spreadsheets are large and reading/loading them can take many seconds.
        # --- Hence, we cache the initial instance of the demographics so that new versions
        # --- don't have re-read the spreadsheets.
        # ------------------------------------------------------------------------------------
        if cls._inital_demog_cache:
            return copy.deepcopy(cls._inital_demog_cache)

        # -------------------------------------------------------
        # --- Initial population has data for different nodes so
        # --- we can get use data from UN World Pop
        # -------------------------------------------------------
        pop_df = pd.DataFrame(NODE_DATA)

        uwp_country = "Zambia"
        uwp_version = "2015"

        total_pop, age_distribution_yar = unwp.extract_population_by_age_and_distribution(country=uwp_country,
                                                                                          version=uwp_version,
                                                                                          year=1960)

        fertility_yar = unwp.extract_fertility(country=uwp_country, version=uwp_version)
        male_mortality_yar = unwp.extract_mortality(country=uwp_country, version=uwp_version, gender="male")
        female_mortality_yar = unwp.extract_mortality(country=uwp_country, version=uwp_version, gender="female")

        # -----------------------------------------------------------------------------------------
        # --- Need a longer fitting interval to flatten natural deaths under the peak in HIV deaths
        # -----------------------------------------------------------------------------------------
        male_mortality_yar = infer.infer_natural_mortality(male_mortality_yar, interval_fit=(1950, 1975))
        female_mortality_yar = infer.infer_natural_mortality(female_mortality_yar, interval_fit=(1950, 1975))

        demog = HIVDemographics.from_year_age_rate_data(pop_df=pop_df,
                                                        age_distribution_yar=age_distribution_yar,
                                                        fertility_yar=fertility_yar,
                                                        male_mortality_yar=male_mortality_yar,
                                                        female_mortality_yar=female_mortality_yar,
                                                        society=None)

        cls._inital_demog_cache = copy.deepcopy(demog)

        return demog

    @classmethod
    def _get_concurrency_parameterized_calls(cls,
                                             relationship_type: str,
                                             risk_group: str,
                                             max_simul_rels_male: float,
                                             max_simul_rels_female: float,
                                             prob_xtra_rel_male: float,
                                             prob_xtra_rel_female: float,
                                             node_ids: Union[List[int], None] = None,
                                             label: str = None) -> List[ParameterizedCall]:
        from emodpy_hiv.demographics.library import set_concurrency_parameters

        parameterized_calls = []
        nhp = {'relationship_type': relationship_type, 'risk_group': risk_group, 'node_ids': node_ids}
        hp = {'max_simul_rels_male': max_simul_rels_male, 'max_simul_rels_female': max_simul_rels_female,
              'prob_xtra_rel_male': prob_xtra_rel_male, 'prob_xtra_rel_female': prob_xtra_rel_female}
        if label is None:
            label = cls.generate_label(relationship_type=relationship_type, risk_group=risk_group, node_ids=node_ids)
        pc = ParameterizedCall(func=set_concurrency_parameters, non_hyperparameters=nhp,
                               hyperparameters=hp,
                               label=label)
        parameterized_calls.append(pc)
        return parameterized_calls

    @classmethod
    def _get_pair_formation_parameterized_calls(cls,
                                                relationship_type: str,
                                                risk_assortivity: float,
                                                formation_rate: float,
                                                node_ids: Union[List[int], None] = None,
                                                label: str = None) -> List[ParameterizedCall]:
        from emodpy_hiv.demographics.library import set_pair_formation_parameters

        parameterized_calls = []
        nhp = {'relationship_type': relationship_type, 'node_ids': node_ids}
        hp = {'risk_assortivity': risk_assortivity, 'formation_rate': formation_rate}
        if label is None:
            label = cls.generate_label(relationship_type=relationship_type, node_ids=node_ids)
        pc = ParameterizedCall(func=set_pair_formation_parameters,
                               non_hyperparameters=nhp,
                               hyperparameters=hp,
                               label=label)
        parameterized_calls.append(pc)
        return parameterized_calls

    @classmethod
    def _get_relationship_parameterized_calls(cls, relationship_type: str,
                                              coital_act_rate: float = None,
                                              condom_usage_min: float = None,
                                              condom_usage_mid: float = None,
                                              condom_usage_max: float = None,
                                              condom_usage_rate: float = None,
                                              duration_scale: float = None,
                                              duration_heterogeneity: float = None,
                                              node_ids: Union[List[int], None] = None,
                                              label: str = None) -> List[ParameterizedCall]:
        from emodpy_hiv.demographics.library import set_relationship_parameters
        parameterized_calls = []
        nhp = {'relationship_type': relationship_type, 'node_ids': node_ids}
        hp = {
            'coital_act_rate': coital_act_rate,
            'condom_usage_min': condom_usage_min,
            'condom_usage_mid': condom_usage_mid,
            'condom_usage_max': condom_usage_max,
            'condom_usage_rate': condom_usage_rate,
            'duration_scale': duration_scale,
            'duration_heterogeneity': duration_heterogeneity
        }

        if label is None:
            label = cls.generate_label(relationship_type=relationship_type, node_ids=node_ids)
        pc = ParameterizedCall(func=set_relationship_parameters,
                               non_hyperparameters=nhp,
                               hyperparameters=hp,
                               label=label)
        parameterized_calls.append(pc)
        return parameterized_calls

    @classmethod
    def _get_individual_property_parameterized_calls(cls) -> List[ParameterizedCall]:
        from emodpy_hiv.demographics.library import set_initial_cascade_state_distribution, \
            set_initial_health_care_accessibility_distribution, set_initial_risk_distribution

        risk_data = []
        risk_data.append({"node_id":  1, "low": 0.255498673 }) # noqa: E202, E241
        risk_data.append({"node_id":  2, "low": 0.330078204 }) # noqa: E202, E241
        risk_data.append({"node_id":  3, "low": 0.121868492 }) # noqa: E202, E241
        risk_data.append({"node_id":  4, "low": 0.221650376 }) # noqa: E202, E241
        risk_data.append({"node_id":  5, "low": 0.395151764 }) # noqa: E202, E241
        risk_data.append({"node_id":  6, "low": 0.095060159 }) # noqa: E202, E241
        risk_data.append({"node_id":  7, "low": 0.288686141 }) # noqa: E202, E241
        risk_data.append({"node_id":  8, "low": 0.446287507 }) # noqa: E202, E241
        risk_data.append({"node_id":  9, "low": 0.378176731 }) # noqa: E202, E241
        risk_data.append({"node_id": 10, "low": 0.283319058 }) # noqa: E202

        parameterized_calls = []
        for data in risk_data:
            hp = {'initial_risk_distribution_low': data["low"]}
            nhp = {'node_ids': [data["node_id"]]}
            pc = ParameterizedCall(func=set_initial_risk_distribution,
                                   non_hyperparameters=nhp,
                                   hyperparameters=hp,
                                   label=NODE_ID_TO_NODE_NAME[data["node_id"]])
            parameterized_calls.append(pc)

        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # Since we need to set separate Risk distributions per node, the following
        # is working around a JSON overlay challene where the list of IP's in a node
        # overrides the the list of IPs in Defaults. EMOD does not allow you to override
        # specific IPs.
        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        distribution = [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        pc = ParameterizedCall(set_initial_cascade_state_distribution,
                               non_hyperparameters={
                                   'cascade_state_distribution': distribution,
                                   'node_ids': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
                               })
        parameterized_calls.append(pc)

        pc = ParameterizedCall(set_initial_health_care_accessibility_distribution,
                               hyperparameters={'initial_accessibility': 0.8},
                               non_hyperparameters={'node_ids': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]})
        parameterized_calls.append(pc)
        return parameterized_calls

    @classmethod
    def get_demographics_parameterized_calls(cls, demographics: HIVDemographics) -> List[ParameterizedCall]:
        parameterized_calls = super().get_demographics_parameterized_calls(demographics=demographics)

        concurrency_parameters = {
            COMMERCIAL: {
                LOW:    {'max_simul_rels_male':  0, 'max_simul_rels_female':  0,     # noqa: E241
                         'prob_xtra_rel_male':   1, 'prob_xtra_rel_female':   1},    # noqa: E241
                MEDIUM: {'max_simul_rels_male':  0, 'max_simul_rels_female':  0,     # noqa: E241
                         'prob_xtra_rel_male':   1, 'prob_xtra_rel_female':   1},    # noqa: E241
                HIGH:   {'max_simul_rels_male': 59, 'max_simul_rels_female': 59,     # noqa: E241
                         'prob_xtra_rel_male':   1, 'prob_xtra_rel_female':   1},    # noqa: E241
            },
            TRANSITORY: {
                LOW:    {'max_simul_rels_male': 1.861181867682,    'max_simul_rels_female': 1.95751180447129,      # noqa: E241
                         'prob_xtra_rel_male':  0.184386490919862, 'prob_xtra_rel_female':  0.0608225887738051},   # noqa: E241
                MEDIUM: {'max_simul_rels_male': 2.48697611966592,  'max_simul_rels_female': 2.38937717475767,      # noqa: E241
                         'prob_xtra_rel_male':  0.287814625075045, 'prob_xtra_rel_female':  0.3116864843045},      # noqa: E241
                HIGH:   {'max_simul_rels_male': 1,                 'max_simul_rels_female': 1,                     # noqa: E241
                         'prob_xtra_rel_male':  1,                 'prob_xtra_rel_female':  1},                    # noqa: E241
            },
            INFORMAL: {
                LOW:    {'max_simul_rels_male': 1.34356628685348,  'max_simul_rels_female': 1.21853602956715,     # noqa: E241
                         'prob_xtra_rel_male':  0.286213622743763, 'prob_xtra_rel_female':  0.289761028337281},   # noqa: E241
                MEDIUM: {'max_simul_rels_male': 2.70230847725607,  'max_simul_rels_female': 2.76579821196633,     # noqa: E241
                         'prob_xtra_rel_male':  0.396354197417793, 'prob_xtra_rel_female':  0.404566634121599},   # noqa: E241
                HIGH:   {'max_simul_rels_male': 1,                 'max_simul_rels_female': 1,                    # noqa: E241
                         'prob_xtra_rel_male':  1,                 'prob_xtra_rel_female':  1},                   # noqa: E241
            },
            MARITAL: {
                LOW:    {'max_simul_rels_male': 1,                'max_simul_rels_female': 1,                # noqa: E241
                         'prob_xtra_rel_male':  0,                'prob_xtra_rel_female':  0},               # noqa: E241
                MEDIUM: {'max_simul_rels_male': 1.86152532283732, 'max_simul_rels_female': 1.7602276432905,  # noqa: E241
                         'prob_xtra_rel_male':  1,                'prob_xtra_rel_female':  1},               # noqa: E241
                HIGH:   {'max_simul_rels_male': 1,                'max_simul_rels_female': 1,                # noqa: E241
                         'prob_xtra_rel_male':  1,                'prob_xtra_rel_female':  1},               # noqa: E241
            },
        }
        for relationship_type, by_risk_group in concurrency_parameters.items():
            for risk_group, parameters in by_risk_group.items():
                pcs = cls._get_concurrency_parameterized_calls(relationship_type=relationship_type,
                                                               risk_group=risk_group,
                                                               **parameters)
                parameterized_calls.extend(pcs)

        pair_formation = {
            # assortivity -1 means a matrix of all 1s
            COMMERCIAL: {'risk_assortivity': -1,                'formation_rate': 0.15                }, # noqa: E241, E202
            TRANSITORY: {'risk_assortivity': 0.159566721712443, 'formation_rate': 0.00135026850353989 }, # noqa: E241, E202
            INFORMAL:   {'risk_assortivity': 0.159566721712443, 'formation_rate': 0.00032670785736624 }, # noqa: E241, E202
            MARITAL:    {'risk_assortivity': 0.159566721712443, 'formation_rate': 0.000151940868159564}, # noqa: E241
        }
        for relationship_type, parameters in pair_formation.items():
            pcs = cls._get_pair_formation_parameterized_calls(relationship_type=relationship_type, **parameters)
            parameterized_calls.extend(pcs)

        # This value is by relationship type and node
        condom_usage_max_by_rel_type_by_node = {
             1: {TRANSITORY: 0.819762848239337,  INFORMAL: 0.203364778711639,  MARITAL: 0.247035322003042, COMMERCIAL: 0.85}, # noqa: E241, E126- Central
             2: {TRANSITORY: 0.397075298252557,  INFORMAL: 0.398597987534942,  MARITAL: 0.247035322003042, COMMERCIAL: 0.85}, # noqa: E241      - Copperbelt
             3: {TRANSITORY: 0.999759873647185,  INFORMAL: 1.0000000000000000, MARITAL: 0.247035322003042, COMMERCIAL: 0.85}, # noqa: E241      - Eastern
             4: {TRANSITORY: 0.929201760347435,  INFORMAL: 0.773399898218325,  MARITAL: 0.247035322003042, COMMERCIAL: 0.85}, # noqa: E241      - Luapula
             5: {TRANSITORY: 0.403033833521129,  INFORMAL: 0.285059471718292,  MARITAL: 0.247035322003042, COMMERCIAL: 0.85}, # noqa: E241      - Lusaka
             6: {TRANSITORY: 1.0000000000000000, INFORMAL: 0.88811421915765,   MARITAL: 0.247035322003042, COMMERCIAL: 0.85}, # noqa: E241      - Muchinga
             7: {TRANSITORY: 0.66571668093682,   INFORMAL: 0.367634610613126,  MARITAL: 0.247035322003042, COMMERCIAL: 0.85}, # noqa: E241      - Northwestern
             8: {TRANSITORY: 0.275068086940464,  INFORMAL: 1.0000000000000000, MARITAL: 0.247035322003042, COMMERCIAL: 0.85}, # noqa: E241      - Northern
             9: {TRANSITORY: 0.636776909282425,  INFORMAL: 0.0000000000000000, MARITAL: 0.247035322003042, COMMERCIAL: 0.85}, # noqa: E241      - Southern
            10: {TRANSITORY: 0.401805672308032,  INFORMAL: 0.387841572157711,  MARITAL: 0.247035322003042, COMMERCIAL: 0.85}  # noqa: E241, E131- Western
        }

        # Other relationship parameters are only by relationship type
        relationship_parameters = {
            TRANSITORY: {'coital_act_rate': 0.33, 'condom_usage_min': 0, 'condom_usage_mid': 2001.29656235789,
                         'condom_usage_rate': 2.36008597080353, 'duration_scale': 0.956774771214,
                         'duration_heterogeneity': 0.833333333},
            INFORMAL:   {'coital_act_rate': 0.33, 'condom_usage_min': 0, 'condom_usage_mid': 2002.789368, # noqa: E241
                         'condom_usage_rate': 0.288440457896589, 'duration_scale': 2.03104913138,
                         'duration_heterogeneity': 0.75},
            MARITAL:    {'coital_act_rate': 0.33, 'condom_usage_min': 0, 'condom_usage_mid': 1994.323885, # noqa: E241
                         'condom_usage_rate': 1.66784081403387, 'duration_scale': 22.154455184937,
                         'duration_heterogeneity': 0.666666667},
            COMMERCIAL: {'coital_act_rate': 0.0027397260273972603, 'condom_usage_min': 0.5, 'condom_usage_mid': 1999.5,
                         'condom_usage_rate': 1, 'duration_scale': 0.01917808219,
                         'duration_heterogeneity': 1}
        }
        # NOTE: The all-node versions of hyperparameters defined here override the per-node-id versions. The reason is
        #  that the all-node hyperparameters are added to the returned list LAST (last defined hyperparameter setting
        #  wins in parameterized call application when >1 hyperparameter affects the same model setting).
        # NOTE: due to how demographics sub-objects are built, if we set 1+ items on e.g. relationship parameters,
        #  we need to set them all to prevent bad things. So we apply all-node values directly to individual nodes,
        #  making the parameters all-node by using an across-node-shared label (multi-use, shared parameters)
        for node_id, condom_usage_max_by_rel_type in condom_usage_max_by_rel_type_by_node.items():
            for relationship_type, condom_usage_max in condom_usage_max_by_rel_type.items():
                node_label = cls.generate_label(relationship_type=relationship_type,
                                                node_name=NODE_ID_TO_NODE_NAME[node_id])
                pcs = cls._get_relationship_parameterized_calls(relationship_type=relationship_type,
                                                                condom_usage_max=condom_usage_max,
                                                                node_ids=[node_id],
                                                                label=node_label)
                parameterized_calls.extend(pcs)

                parameters = relationship_parameters[relationship_type]
                all_nodes_label = cls.generate_label(relationship_type=relationship_type)
                pcs = cls._get_relationship_parameterized_calls(relationship_type=relationship_type,
                                                                **parameters,
                                                                node_ids=[node_id],
                                                                label=all_nodes_label)
                parameterized_calls.extend(pcs)

        # setting initial properties
        pcs = cls._get_individual_property_parameterized_calls()
        parameterized_calls.extend(pcs)

        return parameterized_calls

    @classmethod
    def build_reports(cls, reporters: Reporters) -> Reporters:
        """
        Function that creates reports to be used in the experiment. The function must have Reporters
        object as the parameter and return that object. It is assumed that all the reporters come from the
        reporters that are part of EMOD main code. EMOD also supports reporters as custom plug-in dlls,
        however, not through the current emodpy system.

        Args:
            reporters: The Reporters object to add reports to

        Returns:
            Reporters object
        """
        reporters = super().build_reports(reporters=reporters)

        reporters.add(ReportHIVART(reporters_object=reporters))
        reporters.add(ReportHIVInfection(reporters_object=reporters,
                                         report_filter=ReportFilter(start_year=1980,
                                                                    end_year=2050)))
        reporters.add(ReportHIVMortality(reporters_object=reporters))
        reporters.add(ReportTransmission(reporters_object=reporters))
        reporters.add(ReportRelationshipStart(reporters_object=reporters,
                                              include_hiv_disease_statistics=True,         # default
                                              include_other_relationship_statistics=True,  # default
                                              individual_properties=None,                  # default
                                              report_filter=ReportFilter(start_year=1980,
                                                                         end_year=2050,
                                                                         node_ids=None,                # default
                                                                         min_age_years=None,           # default
                                                                         max_age_years=None,           # default
                                                                         must_have_ip_key_value="",    # default
                                                                         must_have_intervention="")))  # default
        reporters.add(InsetChart(reporters_object=reporters,
                                 has_ip=None,  # default
                                 has_interventions=None,      # default
                                 include_pregnancies=False,   # default
                                 include_coital_acts=True,    # default
                                 event_channels_list=["NewInfectionEvent"]))
        # reporting_period of 182.5 is required for calibration because this will cause EMOD to output a mid-year
        # data timepoint. This reported data is INSTANTANEOUS (current state). It is the current-best-approximation to
        # the year-averaged real-world reference data provided used for calibration (in calibration, we compare the
        # instantaneous, mid-year model output to annual-averaged reference data).
        reporters.add(ReportHIVByAgeAndGender(reporters_object=reporters,
                                              collect_gender_data=True,
                                              collect_age_bins_data=[15, 20, 25, 30, 35, 40, 45, 50],
                                              collect_circumcision_data=True,
                                              collect_hiv_data=True,
                                              collect_hiv_stage_data=False,         # default
                                              collect_on_art_data=True,
                                              collect_ip_data=None,                 # default
                                              collect_intervention_data=["Traditional_MC"],
                                              collect_targeting_config_data=None,   # default
                                              add_transmitters=False,               # default
                                              stratify_infected_by_cd4=False,       # default
                                              event_counter_list=["NewInfectionEvent"],
                                              add_relationships=True,
                                              add_concordant_relationships=False,   # default
                                              reporting_period=182.5,
                                              report_filter=ReportFilter(start_year=1980,
                                                                         end_year=2050)))

        return reporters


class ZambiaForTraining(Zambia):
    """
    A version of Zambia that is used for training purposes.
    It has a smaller population and shorter simulation time.
    """
    @classmethod
    def initialize_config(cls, schema_path: Union[str, Path]) -> ReadOnlyDict:
        config = super().initialize_config(schema_path=schema_path)

        # Set the base population to a small value for training purposes
        # This is to avoid the model running for too long during training.
        config.parameters.x_Base_Population = 0.002

        # Reduce the simulation duration by 10 years to decrease runtime.
        config.parameters.Simulation_Duration = config.parameters.Simulation_Duration - (10 * 365)

        return config



from emod_api import campaign as api_campaign

from emodpy.campaign.base_intervention import IndividualIntervention
from emodpy.campaign.event_coordinator import InterventionDistributorEventCoordinator
from emodpy_hiv.campaign.common import TargetDemographicsConfig, ValueMap, PropertyRestrictions, NChooserTargetedDistributionHIV
from emodpy_hiv.utils.targeting_config import AbstractTargetingConfig

# ported from emodpy/campaign/event_coordinator.py
# The StandardEventCoordinator coordinator class distributes an individual-level or node-level intervention to
# a specified fraction of individuals or nodes within a node set.
from emodpy.campaign.event_coordinator import StandardEventCoordinator


class NChooserEventCoordinatorHIV(InterventionDistributorEventCoordinator):
    """
    The NChooserEventCoordinatorHIV coordinator class distributes individual-level interventions to
    exactly N people of a targeted demographic in HIV simulations. This contrasts with other event coordinators that
    distribute an intervention to a percentage of the population, not to an exact count. This event coordinator is
    similar to the NChooserEventCoordinator for other simulation types, but replaces start and end days in
    TargetedDistribution with start and end years in TargetedDistributionHIV and includes HIV-specific restrictions that
    individuals must have in order to qualify for the intervention.

    Args:
        campaign (api_campaign, required):
            - The campaign object to which the event will be added.
        intervention_list (list[IndividualIntervention], required):
            - A list of individual-level interventions to be distributed.
        targeted_distributions (list[TargetedDistributionHIV], required):
            - A list of TargetedDistributionHIV objects specifying when, to whom, and how many interventions are distributed.
    """
    def __init__(self,
                 campaign: api_campaign,
                 intervention_list: list[IndividualIntervention],
                 targeted_distributions: list[NChooserTargetedDistributionHIV]):
        """
        NChooserEventCoordinatorHIV class to create a
        """
        super().__init__(campaign, "NChooserEventCoordinatorHIV", intervention_list)

        self._coordinator.Distributions = [targeted_distribution.to_schema_dict(campaign)
                                           for targeted_distribution in targeted_distributions]


class ReferenceTrackingEventCoordinatorTrackingConfig(InterventionDistributorEventCoordinator):
    """
    The ReferenceTrackingEventCoordinatorTrackingConfig coordinator class defines a particular prevalence of an
    individual-level attribute that should be present in a population over time, and a corresponding intervention that
    will cause individuals to acquire that attribute. The coordinator tracks the actual prevalence of that attribute
    against the desired prevalence; it will poll the population of nodes it has been assigned to determine how many
    people have the attribute. When coverage is less than the desired prevalence, it will distribute enough of the
    designated intervention to reach the desired prevalence. This coordinator is similar to the
    ReferenceTrackingEventCoordinator, but allows an attribute in the population to be polled, not only the intervention
    itself having been received. This allows for tracking overall coverage when, potentially, multiple routes exist for
    individuals to have acquired the same target attribute.

    Args:
        campaign (api_campaign, required):
            - The campaign object to which the event will be added.
        intervention_list (list[IndividualIntervention], required):
            - A list of individual-level interventions to be distributed.
        time_value_map (ValueMap, required):
            - A ValueMap object to map coverages over a defined range of time.
        tracking_config (AbstractTargetingConfig, required):
            - An AbstractTargetingConfig to define the attribute to be tracked within the targeted group.
            - The number of people that have this attribute is the numerator while the other targeting parameters define the denominator of the coverage.
            - The intervention will be distributed to people without the attribute, if coverage is below the target level the time of polling.
        end_year (float, optional):
            - The year to stop distributing the intervention.
            - Defines the time period to distribute the intervention along with start_year. The intervention is evenly distributed between each time step in the time period.
            - Minimum value: 1900,
            - Maximum value: 2200,
            - Default value: 2200.
        update_period (float, optional):
            - The time period in days between updates to the intervention distribution.
            - Minimum value: 1,
            - Maximum value: 3650,
            - Default value: 365.
        target_demographics_config (TargetDemographicsConfig, optional):
            - A TargetDemographicsConfig to define the demographic characteristics of individuals targeted by the intervention.
            - Please refer to the emodpy_hiv.campaign.common.TargetDemographicsConfig module for more information.
            - Please note that the demographic_coverage is not used in this coordinator. Use TargetDemographicsConfig(demographic_coverage=None) to avoid errors.
            - Default value: TargetDemographicsConfig(demographic_coverage=None).
        property_restrictions (PropertyRestrictions, optional):
            - A PropertyRestrictions to define the individual-level or node-level property restrictions in the coordinator.
            - Default value: None.
        targeting_config (AbstractTargetingConfig, optional):
            - An AbstractTargetingConfig to define who to target individuals with the intervention besides the target_demographics_config and property_restrictions.
            - Please refer to the emodpy_hiv.utils.targeting_config module for more information.
            - Default value: None.
    """
    def __init__(self,
                 campaign: api_campaign,
                 intervention_list: list[IndividualIntervention],
                 time_value_map: ValueMap,
                 targeting_config: AbstractTargetingConfig,
                 tracking_config: AbstractTargetingConfig,
                 end_year: float = 2200,
                 update_period: float = 365,
                 target_demographics_config: TargetDemographicsConfig = TargetDemographicsConfig(demographic_coverage=None),
                 property_restrictions: PropertyRestrictions = None):

        super().__init__(campaign, "ReferenceTrackingEventCoordinatorTrackingConfig", intervention_list)

        if not all(isinstance(intervention, IndividualIntervention) for intervention in intervention_list):
            raise ValueError("intervention_list should contain only IndividualIntervention objects.")

        if target_demographics_config is not None:
            if target_demographics_config.demographic_coverage is not None:
                raise ValueError("The demographic_coverage parameter is not used in this coordinator. "
                                 "Use TargetDemographicsConfig(demographic_coverage=None) to avoid errors.")
            target_demographics_config._set_target_demographics(self._coordinator)
        if property_restrictions is not None:
            property_restrictions._set_property_restrictions(self._coordinator)
        if targeting_config is not None:
            self._coordinator.Targeting_Config = targeting_config.to_schema_dict(campaign)

        self._coordinator.End_Year = end_year
        self._coordinator.Update_Period = update_period
        self._coordinator.Time_Value_Map = time_value_map.to_schema_dict(campaign)
        self._coordinator.Tracking_Config = tracking_config.to_simple_dict(campaign)


# __all_exports: A list of classes that are intended to be exported from this module.
__all_exports = [StandardEventCoordinator, NChooserEventCoordinatorHIV,
                 ReferenceTrackingEventCoordinatorTrackingConfig]

# The following loop sets the __module__ attribute of each class in __all_exports to the name of the current module.
# This is done to ensure that when these classes are imported from this module, their __module__ attribute correctly
# reflects their source module.
# During the documentation build with Sphinx, these classes will be displayed as belonging to the 'emodpy_hiv' package,
# not the 'emodpy' package.
# For example, the 'PropertyRestrictions' class will be documented as 'emodpy_hiv.campaign.common.PropertyRestrictions(...)'.
# This is crucial for accurately representing the source of these classes in the documentation.

for _ in __all_exports:
    _.__module__ = __name__

# __all__: A list that defines the public interface of this module.
# This is essential to ensure that Sphinx builds documentation for these classes, including those that are imported
# from emodpy.
# It contains the names of all the classes that should be accessible when this module is imported using the syntax
# 'from module import *'.
# Here, it is set to the names of all classes in __all_exports.

__all__ = [_.__name__ for _ in __all_exports]



import sys
import argparse
import pandas as pd
import json

import emodpy_hiv.plotting.helpers as helpers
import emodpy_hiv.plotting.plot_inset_chart as pic
from emodpy_hiv.plotting.channel_report import ChannelReport


def calculate_mean(dir_name: str):
    """
    Calculate the mean of the InsetChart.json files in the given directory.

    Args:
        dir_name:
            Directory with InsetChart.json files

    Return:
        mean_cr: Mean ChannelReport object
        raw_data_list: List of raw data dictionaries from the InsetChart.json files
    """
    test_filenames = helpers.get_filenames(dir_or_filename=dir_name,
                                           file_prefix="InsetChart",
                                           file_extension="json")

    raw_data_list = []
    total_df = pd.DataFrame()
    for test_fn in test_filenames:
        with open(test_fn, "r") as test_file:
            test_json = json.loads(test_file.read())
            raw_data_list.append(test_json)
        df = ChannelReport.convert_to_df(test_json)
        total_df = pd.concat([total_df, df])

    mean_df = total_df.groupby("Time").mean().reset_index()
    mean_cr = ChannelReport(df=mean_df)
    mean_cr = mean_cr.convert_df_to_channel_report(df=mean_df)

    return mean_cr, raw_data_list


def plot_mean(dir1: str,
              dir2: str,
              dir3: str,
              title: str = None,
              show_raw_data: bool = False,
              subplot_index_min: int = 0,
              subplot_index_max: int = 100,
              output: str = None):
    """
    Plot the mean of the InsetChart.json files in the given directories.

    Args:
        dir1:
            Directory with InsetChart.json files

        dir2:
            Directory with InsetChart.json files

        dir3:
            Directory with InsetChart.json files

        title:
            Title of Plot

        show_raw_data:
            If true, shows the raw/individual simulation data in a lighter color.

        subplot_index_min:
            The index of the first subplot to show based on the alphabetical
            order of the channels in the report.

        subplot_index_min:
            The index of the last subplot to show based on the alphabetical
            order of the channels in the report.

        output:
            If provided, a directory will be created and images saved to the folder.  If not provided, it opens windows.

    Return:
        None, if output is provided, an image will be saved to the output directory, else a window will be opened.
    """
    dir_names = []
    mean_data = []
    raw_data_lists = []

    if dir1 is not None:
        dir_names.append(dir1)
        mean_cr_1, raw_data_list_1 = calculate_mean(dir1)
        mean_data.append(mean_cr_1)
        raw_data_lists.append(raw_data_list_1)

    if dir2 is not None:
        dir_names.append(dir2)
        mean_cr_2, raw_data_list_2 = calculate_mean(dir2)
        mean_data.append(mean_cr_2)
        raw_data_lists.append(raw_data_list_2)

    if dir3 is not None:
        dir_names.append(dir3)
        mean_cr_3, raw_data_list_3 = calculate_mean(dir3)
        mean_data.append(mean_cr_3)
        raw_data_lists.append(raw_data_list_3)

    plot_name = title
    if plot_name is None:
        plot_name = "InsetChart_Compare"

    if title is None:
        title = "" + pic.create_title_string(reference=None, data_filenames=dir_names)

    if not show_raw_data:
        raw_data_lists = None

    pic.plot_data(title=title,
                  ref_data=None,
                  test_data=mean_data,
                  raw_data_list_of_lists=raw_data_lists,
                  test_filenames=dir_names,
                  subplot_index_min=subplot_index_min,
                  subplot_index_max=subplot_index_max,
                  img_dir=output,
                  plot_name=plot_name)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('dir1', default=None, nargs='?', help='Directory, or parent directory that contains subdirectories, of InsetChart.json files')
    parser.add_argument('dir2', default=None, nargs='?', help='Directory, or parent directory that contains subdirectories, of InsetChart.json files')
    parser.add_argument('dir3', default=None, nargs='?', help='Directory, or parent directory that contains subdirectories, of InsetChart.json files')
    parser.add_argument('-r', '--raw', help='If provided, shows the raw/individual simulation data in a lighter color.', action='store_true')
    parser.add_argument('-m', '--subplot_index_min', default=0, type=int, help='Index of the first subplot to show.')
    parser.add_argument('-x', '--subplot_index_max', default=100, type=int, help='Index of the last subplot to show.')
    parser.add_argument('-t', '--title', default=None, help='Title of Plot')
    parser.add_argument('-o', '--output', default=None, help='If provided, a directory will be created and images saved to the folder.  If not provided, it opens windows.')

    args = parser.parse_args()

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit()

    if args.subplot_index_min >= args.subplot_index_max:
        print("Error: subplot_index_min must be less than subplot_index_max.")
        sys.exit()

    plot_mean(dir1=args.dir1,
              dir2=args.dir2,
              dir3=args.dir3,
              title=args.title,
              show_raw_data=args.raw,
              subplot_index_min=args.subplot_index_min,
              subplot_index_max=args.subplot_index_max,
              output=args.output)



from typing import List

from emodpy_hiv.demographics.hiv_demographics import HIVDemographics


def set_concurrency_parameters(demographics: HIVDemographics,
                               relationship_type: str,
                               risk_group: str,
                               max_simul_rels_male: float = None,
                               max_simul_rels_female: float = None,
                               prob_xtra_rel_male: float = None,
                               prob_xtra_rel_female: float = None,
                               node_ids: List[int] = None) -> None:

    demographics.set_concurrency_params_by_type_and_risk(relationship_type=relationship_type,
                                                         risk_group=risk_group,
                                                         max_simul_rels_male=max_simul_rels_male,
                                                         max_simul_rels_female=max_simul_rels_female,
                                                         prob_xtra_rel_male=prob_xtra_rel_male,
                                                         prob_xtra_rel_female=prob_xtra_rel_female,
                                                         node_ids=node_ids)


def set_pair_formation_parameters(demographics: HIVDemographics,
                                  relationship_type: str,
                                  formation_rate: float = None,
                                  risk_assortivity: float = None,
                                  node_ids: List[int] = None) -> None:
    def _ra_matrix(v: float):
        if v is None:
            matrix = None
        elif v == -1:
            matrix = [[1, 1, 1], [1, 1, 1], [1, 1, 1]]
        else:
            matrix = [[v, 1 - v, 0], [1 - v, v, v], [0, v, 1 - v]]
        return matrix

    assortivity_matrix = _ra_matrix(v=risk_assortivity)
    demographics.set_pair_formation_parameters(relationship_type=relationship_type,
                                               formation_rate=formation_rate,
                                               assortivity_matrix=assortivity_matrix,
                                               node_ids=node_ids)


def set_relationship_parameters(demographics: HIVDemographics,
                                relationship_type: str,
                                coital_act_rate: float = None,
                                condom_usage_min: float = None,
                                condom_usage_mid: float = None,
                                condom_usage_max: float = None,
                                condom_usage_rate: float = None,
                                duration_scale: float = None,
                                duration_heterogeneity: float = None,
                                node_ids: List[int] = None) -> None:
    demographics.set_relationship_parameters(relationship_type=relationship_type,
                                             coital_act_rate=coital_act_rate,
                                             condom_usage_min=condom_usage_min,
                                             condom_usage_mid=condom_usage_mid,
                                             condom_usage_max=condom_usage_max,
                                             condom_usage_rate=condom_usage_rate,
                                             duration_scale=duration_scale,
                                             duration_heterogeneity=duration_heterogeneity,
                                             node_ids=node_ids)


#
# INITIAL IP DISTRIBUTION METHODS
#


def set_initial_risk_distribution(demographics: HIVDemographics, initial_risk_distribution_low: float,
                                  node_ids: List[int] = None) -> None:
    def _risk_distribution(low: float):
        return None if low is None else [low, 1 - low, 0]

    distribution = _risk_distribution(low=initial_risk_distribution_low)
    demographics.add_or_update_initial_risk_distribution(distribution=distribution, node_ids=node_ids)


def set_initial_cascade_state_distribution(demographics: HIVDemographics,
                                           cascade_state_distribution: List[float],
                                           node_ids: List[int] = None) -> None:
    demographics.add_or_update_initial_cascade_state_distribution(distribution=cascade_state_distribution,
                                                                  node_ids=node_ids)


def set_initial_health_care_accessibility_distribution(demographics: HIVDemographics,
                                                       initial_accessibility: float,
                                                       node_ids: List[int] = None) -> None:
    distribution = [initial_accessibility, 1 - initial_accessibility]
    demographics.add_or_update_initial_health_care_accessibility_distribution(distribution=distribution,
                                                                              node_ids=node_ids)
